from fontTools.misc.transform import Identity

from ufo2ft.filters.decomposeComponents import (
    DecomposeComponentsFilter,
    DecomposeComponentsIFilter,
)

IDENTITY_2x2 = Identity[:4]


def _isTransformed(component):
    return component.transformation[:4] != IDENTITY_2x2


class DecomposeTransformedComponentsFilter(DecomposeComponentsFilter):
    def filter(self, glyph):
        if any(_isTransformed(c) for c in glyph.components):
            return super().filter(glyph)
        return False


class DecomposeTransformedComponentsIFilter(DecomposeComponentsIFilter):
    def filter(self, glyphName, glyphs):
        # We decomposes the glyph in *all* masters if any contains a transformed
        # component: https://github.com/googlefonts/ufo2ft/issues/507
        if not any(any(_isTransformed(c) for c in g.components) for g in glyphs):
            return False
        return super().filter(glyphName, glyphs)



"""Alternative implementation of KernFeatureWriter.

This behaves like the primary kern feature writer, with the important difference
of grouping kerning data into lookups by kerning direction, not script, like the
feature writer in ufo2ft v2.30 and older did.

The original idea for the primary splitter was to generate smaller, easier to
pack lookups for each script exclusively, as cross-script kerning dos not work
in browsers. However, other applications may allow it, e.g. Adobe's InDesign.
Subsequently, it was modified to clump together lookups that cross-reference
each other's scripts, negating the size advantages if you design fonts with
cross-script kerning for designer ease.

As a special edge case, InDesign's default text shaper does not properly itemize
text runs, meaning it may group different scripts into the same run unless the
user specifically marks some text as being a specific script or language. To
make all kerning reachable in that case, it must be put into a single broad LTR,
RTL or neutral direction lookup instead of finer script clusters. That will make
it work in all cases, including when there is no cross-script kerning to fuse
different lookups together.

Testing showed that size benefits are clawed back with the use of the HarfBuzz
repacker (during compilation) and GPOS compression (after compilation) at
acceptable speed.
"""

from __future__ import annotations

import enum
import itertools
import logging
import sys
from collections import OrderedDict
from types import SimpleNamespace
from typing import Any, Iterator, Mapping, cast

import fontTools.feaLib.ast as fea_ast
from fontTools import unicodedata
from fontTools.unicodedata import script_horizontal_direction

from ufo2ft.featureWriters import BaseFeatureWriter, ast
from ufo2ft.util import (
    DFLT_SCRIPTS,
    classifyGlyphs,
    describe_ufo,
    quantize,
)

from .kernFeatureWriter import (
    AMBIGUOUS_BIDIS,
    DIST_ENABLED_SCRIPTS,
    LTR_BIDI_TYPES,
    RTL_BIDI_TYPES,
    KerningPair,
    addClassDefinition,
    collectKerningGroups,
    getVariableKerningPairs,
)

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias

LOGGER = logging.getLogger(__name__)

KerningGroup: TypeAlias = "Mapping[str, tuple[str, ...]]"


class Direction(enum.Enum):
    Neutral = "dflt"
    LeftToRight = "ltr"
    RightToLeft = "rtl"

    def __lt__(self, other: Direction) -> bool:
        if not isinstance(other, Direction):
            return NotImplemented

        return self.name < other.name


class KernContext(SimpleNamespace):
    bidiGlyphs: dict[Direction, set[str]]
    compiler: Any
    default_source: Any
    existingFeatures: Any
    feaFile: Any
    feaLanguagesByTag: dict[str, list[str]]
    font: Any
    gdefClasses: Any
    glyphBidi: dict[str, set[Direction]]
    glyphDirection: dict[str, set[Direction]]
    glyphSet: OrderedDict[str, Any]
    insertComments: Any
    isVariable: bool
    kerning: Any
    knownScripts: set[str]
    side1Membership: dict[str, str]
    side2Membership: dict[str, str]
    todo: Any


class KernFeatureWriter(BaseFeatureWriter):
    """Generates a kerning feature based on groups and rules contained
    in an UFO's kerning data.

    If the `quantization` argument is given in the filter options, the resulting
    anchors are rounded to the nearest multiple of the quantization value.

    ## Implementation Notes

    The algorithm works like this:

    * Parse GDEF GlyphClassDefinition from UFO features.fea to get the set of
      "Mark" glyphs (this will be used later to decide whether to add
      ignoreMarks flag to kern lookups containing pairs between base and mark
      glyphs).
    * Get the ordered glyphset for the font, for filtering kerning groups and
      kernings that reference unknown glyphs.
    * Determine which scripts the kerning affects (read: "the font most probably
      supports"), to know which lookups to generate later:
        * First, determine the unambiguous script associations for each
          (Unicoded) glyph in the glyphset, as in, glyphs that have a single
          entry for their Unicode script extensions property;
        * then, parse the `languagesystem` statements in the provided feature
          file to add on top.
    * Compile a Unicode cmap from the UFO and a GSUB table from the features so
      far, so we can determine the bidirectionality class, so we can later
      filter out kerning pairs that would mix RTL and LTR glyphs, which will not
      occur in applications, and put the pairs into their correct lookup.
      Unicode BiDi classes R and AL are considered R. Common characters and
      numbers are considered neutral even when their BiDi class says otherwise,
      so they'll end up in the common lookup available to all scripts.
    * Get the kerning groups from the UFO and filter out glyphs not in the
      glyphset and empty groups. Remember which group a glyph is a member of,
      for kern1 and kern2, so we can later reconstruct per-direction groups.
    * Get the bare kerning pairs from the UFO, filtering out pairs with unknown
      groups or glyphs not in the glyphset and (redundant) zero class-to-class
      kernings and optionally quantizing kerning values.
    * Optionally, split kerning pairs into base (only base against base) and
      mark (mixed mark and base) pairs, according to the glyphs' GDEF category,
      so that kerning against marks can be accounted for correctly later.
    * Go through all kerning pairs and split them up by direction, to put them
      in different lookups. In pairs with common glyphs, assume the direction of
      the dominant script, in pairs of common glyphs, assume no direction. Pairs
      with clashing script directions are dropped.
        * Partition the first and second side of a pair by BiDi direction (as
          above) and emit only those with the same direction or a strong
          direction and neutral one.
        * Discard pairs that mix RTL and LTR BiDi types, because they won't show
          up in applications due to how Unicode text is split into runs.
        * Glyphs will have only one direction assigned to them. * Preserve the
          type of the kerning pair, so class-to-class kerning stays that way,
          even when there's only one glyph on each side.
    * Reconstruct kerning group names for the newly split classes. This is done
      for debuggability; it makes no difference for the final font binary.
        * This first looks at the neutral lookups and then all others, assigning
          new group names are it goes. A class like `@kern1.something = [foo bar
          baz]` may be split up into `@kern1.dflt.something = [foo]` and
          `@kern1.ltr.something = [bar baz]`. Note: If there is no dedicated
          dflt lookup, common glyph classes like `[foo]` might carry the name
          `@kern1.ltr.foo` if the class was first encountered while going over
          the ltr lookup.
    * Make a `kern` (and potentially `dist`) feature block and register the
      lookups for each script. Some scripts need to be registered in the `dist`
      feature for some shapers to discover them, e.g. Yezi.
    * Write the new glyph class definitions and then the lookups and feature
      blocks to the feature file.
    """

    tableTag = "GPOS"
    features = frozenset(["kern", "dist"])
    options = dict(ignoreMarks=True, quantization=1)

    def setContext(self, font, feaFile, compiler=None):
        ctx: KernContext = cast(
            KernContext, super().setContext(font, feaFile, compiler=compiler)
        )

        if hasattr(font, "findDefault"):
            ctx.default_source = font.findDefault().font
        else:
            ctx.default_source = font

        # Unless we use the legacy append mode (which ignores insertion
        # markers), if the font (Designspace: default source) contains kerning
        # and the feaFile contains `kern` or `dist` feature blocks, but we have
        # no insertion markers (or they were misspelt and ignored), warn the
        # user that the kerning blocks in the feaFile take precedence and other
        # kerning is dropped.
        if (
            self.mode == "skip"
            and ctx.default_source.kerning
            and ctx.existingFeatures & self.features
            and not ctx.insertComments
        ):
            LOGGER.warning(
                "%s: font has kerning, but also manually written kerning features "
                "without an insertion comment. Dropping the former.",
                describe_ufo(ctx.default_source),
            )

        # Remember which languages are defined for which OT tag, as all
        # generated kerning needs to be registered for the script's `dflt`
        # language, but also all those the designer defined manually. Otherwise,
        # setting any language for a script would deactivate kerning.
        feaLanguagesByScript = ast.getScriptLanguageSystems(feaFile, excludeDflt=False)
        ctx.feaLanguagesByTag = {
            otTag: languages
            for _, languageSystems in feaLanguagesByScript.items()
            for otTag, languages in languageSystems
        }

        ctx.glyphSet = self.getOrderedGlyphSet()
        ctx.gdefClasses = self.getGDEFGlyphClasses()
        ctx.knownScripts = self.guessFontScripts()

        # We need the direction of a glyph (with common characters considered
        # neutral or "dflt") to know in which of the three lookups to put the
        # pair.
        cmap = self.makeUnicodeToGlyphNameMapping()
        gsub = self.compileGSUB()
        extras = self.extraSubstitutions()
        dirGlyphs = classifyGlyphs(unicodeScriptDirection, cmap, gsub, extras)
        neutral_glyphs = (
            ctx.glyphSet.keys()
            - dirGlyphs.get(Direction.LeftToRight, set())
            - dirGlyphs.get(Direction.RightToLeft, set())
        )
        dirGlyphs[Direction.Neutral] = neutral_glyphs
        glyphDirection = {}
        for direction, glyphs in dirGlyphs.items():
            for name in glyphs:
                glyphDirection.setdefault(name, set()).add(direction)
        ctx.glyphDirection = glyphDirection

        # We need the BiDi class of a glyph to reject kerning of RTL glyphs
        # against LTR glyphs.
        ctx.bidiGlyphs = classifyGlyphs(unicodeBidiType, cmap, gsub, extras)
        neutral_glyphs = (
            ctx.glyphSet.keys()
            - ctx.bidiGlyphs.get(Direction.LeftToRight, set())
            - ctx.bidiGlyphs.get(Direction.RightToLeft, set())
        )
        ctx.bidiGlyphs[Direction.Neutral] = neutral_glyphs
        glyphBidi = {}
        for direction, glyphs in ctx.bidiGlyphs.items():
            for name in glyphs:
                glyphBidi.setdefault(name, set()).add(direction)
        ctx.glyphBidi = glyphBidi

        ctx.kerning = extract_kerning_data(ctx, cast(SimpleNamespace, self.options))

        return ctx

    def shouldContinue(self):
        if (
            not self.context.kerning.base_pairs_by_direction
            and not self.context.kerning.mark_pairs_by_direction
        ):
            self.log.debug("No kerning data; skipped")
            return False

        return super().shouldContinue()

    def _write(self):
        self.context: KernContext
        self.options: SimpleNamespace
        lookups = make_kerning_lookups(self.context, self.options)
        if not lookups:
            self.log.debug("kerning lookups empty; skipped")
            return False

        features = make_feature_blocks(self.context, lookups)
        if not features:
            self.log.debug("kerning features empty; skipped")
            return False

        # extend feature file with the new generated statements
        feaFile = self.context.feaFile

        # first add the glyph class definitions
        classDefs = self.context.kerning.classDefs
        newClassDefs = [c for _, c in sorted(classDefs.items())]

        featureBlocks = [features[tag] for tag in ["kern", "dist"] if tag in features]

        # Collect only the lookups that are referenced by the features we're
        # writing, to avoid inserting unreferenced/dangling lookups when a
        # feature block already exists in the feature file.
        referencedLookups = {
            statement.lookup
            for feature in featureBlocks
            for statement in feature.statements
            if isinstance(statement, ast.LookupReferenceStatement)
        }

        lookupGroups = []
        for _, lookupGroup in sorted(lookups.items(), key=lambda x: x[0].value):
            lookupGroups.extend(
                lkp
                for lkp in lookupGroup.values()
                if lkp in referencedLookups and lkp not in lookupGroups
            )

        self._insert(
            feaFile=feaFile,
            classDefs=newClassDefs,
            lookups=lookupGroups,
            features=featureBlocks,
        )
        return True


def unicodeBidiType(uv: int) -> Direction | None:
    """Return Direction.RightToLeft for characters with strong RTL
    direction, or Direction.LeftToRight for strong LTR and European and Arabic
    numbers, or None for neutral direction.
    """
    bidiType = unicodedata.bidirectional(chr(uv))
    if bidiType in RTL_BIDI_TYPES:
        return Direction.RightToLeft
    elif bidiType in LTR_BIDI_TYPES:
        return Direction.LeftToRight
    return None


def unicodeScriptDirection(uv: int) -> Direction | None:
    script = unicodedata.script(chr(uv))
    if script in DFLT_SCRIPTS:
        return None
    direction = unicodedata.script_horizontal_direction(script, "LTR")
    if direction == "LTR":
        return Direction.LeftToRight
    elif direction == "RTL":
        return Direction.RightToLeft
    raise ValueError(f"Unknown direction {direction}")


def extract_kerning_data(context: KernContext, options: SimpleNamespace) -> Any:
    side1Groups, side2Groups = get_kerning_groups(context)
    if context.isVariable:
        pairs = get_variable_kerning_pairs(context, options)
    else:
        pairs = get_kerning_pairs(context, options, side1Groups, side2Groups)

    if options.ignoreMarks:
        marks = context.gdefClasses.mark
        base_pairs, mark_pairs = split_base_and_mark_pairs(pairs, marks)
    else:
        base_pairs = pairs
        mark_pairs = []

    base_pairs_by_direction = split_kerning(context, base_pairs)
    mark_pairs_by_direction = split_kerning(context, mark_pairs)

    return SimpleNamespace(
        base_pairs_by_direction=base_pairs_by_direction,
        mark_pairs_by_direction=mark_pairs_by_direction,
        side1Classes={},
        side2Classes={},
        classDefs={},
    )


def get_kerning_groups(context: KernContext) -> tuple[KerningGroup, KerningGroup]:
    side1Groups, side2Groups, side1Membership, side2Membership = collectKerningGroups(
        context.font, context.glyphSet, context.isVariable
    )
    context.side1Membership = side1Membership
    context.side2Membership = side2Membership
    return side1Groups, side2Groups


def get_kerning_pairs(
    context: KernContext,
    options: SimpleNamespace,
    side1Classes: KerningGroup,
    side2Classes: KerningGroup,
) -> list[KerningPair]:
    glyphSet = context.glyphSet
    font = context.font
    kerning: Mapping[tuple[str, str], float] = font.kerning
    quantization = options.quantization

    result = []
    for (side1, side2), value in kerning.items():
        firstIsClass, secondIsClass = (side1 in side1Classes, side2 in side2Classes)
        # Filter out pairs that reference missing groups or glyphs.
        if not firstIsClass and side1 not in glyphSet:
            continue
        if not secondIsClass and side2 not in glyphSet:
            continue
        # Ignore zero-valued class kern pairs. They are the most general
        # kerns, so they don't override anything else like glyph kerns would
        # and zero is the default.
        if firstIsClass and secondIsClass and value == 0:
            continue
        if firstIsClass:
            side1 = side1Classes[side1]
        if secondIsClass:
            side2 = side2Classes[side2]
        value = quantize(value, quantization)
        result.append(KerningPair(side1, side2, value))

    return result


def get_variable_kerning_pairs(
    context: KernContext,
    options: SimpleNamespace,
) -> list[KerningPair]:
    # This legacy writer shares the default KernFeatureWriter's variable kerning
    # extraction, so the logic lives in exactly one place. The default writer
    # passes the designspace and glyph set explicitly rather than via the
    # KernContext, so unpack them here.
    return getVariableKerningPairs(
        context.font,
        context.glyphSet,
        options,
    )


def split_base_and_mark_pairs(
    pairs: list[KerningPair], marks: set[str]
) -> tuple[list[KerningPair], list[KerningPair]]:
    if not marks:
        return list(pairs), []

    basePairs: list[KerningPair] = []
    markPairs: list[KerningPair] = []
    for pair in pairs:
        # Disentangle kerning between bases and marks by splitting a pair
        # into a list of base-to-base pairs (basePairs) and a list of
        # base-to-mark, mark-to-base and mark-to-mark pairs (markPairs).
        # This ensures that "kerning exceptions" (a kerning pair modifying
        # the effect of another) work as intended because these related
        # pairs end up in the same list together.
        side1Bases: tuple[str, ...] | str | None = None
        side1Marks: tuple[str, ...] | str | None = None
        if pair.firstIsClass:
            side1Bases = tuple(glyph for glyph in pair.side1 if glyph not in marks)
            side1Marks = tuple(glyph for glyph in pair.side1 if glyph in marks)
        elif pair.side1 in marks:
            side1Marks = pair.side1
        else:
            side1Bases = pair.side1

        side2Bases: tuple[str, ...] | str | None = None
        side2Marks: tuple[str, ...] | str | None = None
        if pair.secondIsClass:
            side2Bases = tuple(glyph for glyph in pair.side2 if glyph not in marks)
            side2Marks = tuple(glyph for glyph in pair.side2 if glyph in marks)
        elif pair.side2 in marks:
            side2Marks = pair.side2
        else:
            side2Bases = pair.side2

        if side1Bases and side2Bases:  # base-to-base
            basePairs.append(KerningPair(side1Bases, side2Bases, value=pair.value))

        if side1Bases and side2Marks:  # base-to-mark
            markPairs.append(KerningPair(side1Bases, side2Marks, value=pair.value))
        if side1Marks and side2Bases:  # mark-to-base
            markPairs.append(KerningPair(side1Marks, side2Bases, value=pair.value))
        if side1Marks and side2Marks:  # mark-to-mark
            markPairs.append(KerningPair(side1Marks, side2Marks, value=pair.value))

    return basePairs, markPairs


def split_kerning(
    context: KernContext,
    pairs: list[KerningPair],
) -> dict[Direction, list[KerningPair]]:
    # Split kerning into per-direction buckets, so we can drop them into their
    # own lookups.
    glyph_bidi = context.glyphBidi
    glyph_direction = context.glyphDirection
    kerning_per_direction: dict[Direction, list[KerningPair]] = {}
    for pair in pairs:
        for direction, split_pair in partition_by_direction(
            pair, glyph_bidi, glyph_direction
        ):
            kerning_per_direction.setdefault(direction, []).append(split_pair)

    for pairs in kerning_per_direction.values():
        pairs.sort()

    return kerning_per_direction


def partition_by_direction(
    pair: KerningPair,
    glyph_bidi: Mapping[str, set[Direction]],
    glyph_direction: Mapping[str, set[Direction]],
) -> Iterator[tuple[Direction, KerningPair]]:
    """Split a potentially mixed-direction pair into pairs of the same
    or compatible direction."""

    side1Bidis: dict[Direction, set[str]] = {}
    side2Bidis: dict[Direction, set[str]] = {}
    side1Directions: dict[Direction, set[str]] = {}
    side2Directions: dict[Direction, set[str]] = {}
    for glyph in pair.firstGlyphs:
        bidis = glyph_bidi[glyph]
        directions = glyph_direction[glyph]
        for bidi in bidis:
            side1Bidis.setdefault(bidi, set()).add(glyph)
        for direction in directions:
            side1Directions.setdefault(direction, set()).add(glyph)
    for glyph in pair.secondGlyphs:
        bidis = glyph_bidi[glyph]
        directions = glyph_direction[glyph]
        for bidi in bidis:
            side2Bidis.setdefault(bidi, set()).add(glyph)
        for direction in directions:
            side2Directions.setdefault(direction, set()).add(glyph)

    for side1Direction, side2Direction in itertools.product(
        sorted(side1Directions), sorted(side2Directions)
    ):
        localSide1: str | tuple[str, ...]
        localSide2: str | tuple[str, ...]
        if pair.firstIsClass:
            localSide1 = tuple(sorted(side1Directions[side1Direction]))
        else:
            assert len(side1Directions[side1Direction]) == 1
            (localSide1,) = side1Directions[side1Direction]
        if pair.secondIsClass:
            localSide2 = tuple(sorted(side2Directions[side2Direction]))
        else:
            assert len(side2Directions[side2Direction]) == 1
            (localSide2,) = side2Directions[side2Direction]

        # Skip pairs with clashing directions (e.g. "a" to "alef-ar").
        if side1Direction != side2Direction and not any(
            side is Direction.Neutral for side in (side1Direction, side2Direction)
        ):
            LOGGER.info(
                "Skipping part of a kerning pair <%s %s %s> with mixed direction (%s, %s)",
                localSide1,
                localSide2,
                pair.value,
                side1Direction.name,
                side2Direction.name,
            )
            continue

        # Skip pairs with clashing BiDi classes (e.g. "alef-ar" to "one-ar").
        localSide1Bidis = {
            bidi
            for glyph in side1Directions[side1Direction]
            for bidi in glyph_bidi[glyph]
        }
        localSide2Bidis = {
            bidi
            for glyph in side2Directions[side2Direction]
            for bidi in glyph_bidi[glyph]
        }
        if localSide1Bidis != localSide2Bidis and not any(
            Direction.Neutral in side for side in (localSide1Bidis, localSide2Bidis)
        ):
            LOGGER.info(
                "Skipping part of a kerning pair <%s %s %s> with conflicting BiDi classes",
                localSide1,
                localSide2,
                pair.value,
            )
            continue

        dominant_direction = (
            side1Direction if side2Direction is Direction.Neutral else side2Direction
        )
        yield (dominant_direction, KerningPair(localSide1, localSide2, pair.value))


def make_kerning_lookups(
    context: KernContext, options: SimpleNamespace
) -> dict[Direction, dict[str, fea_ast.LookupBlock]]:
    lookups: dict[Direction, dict[str, fea_ast.LookupBlock]] = {}
    if context.kerning.base_pairs_by_direction:
        make_split_kerning_lookups(
            context, options, lookups, context.kerning.base_pairs_by_direction
        )
    if context.kerning.mark_pairs_by_direction:
        make_split_kerning_lookups(
            context,
            options,
            lookups,
            context.kerning.mark_pairs_by_direction,
            ignoreMarks=False,
            suffix="_marks",
        )
    return lookups


def make_split_kerning_lookups(
    context: KernContext,
    options: SimpleNamespace,
    lookups: dict[Direction, dict[str, fea_ast.LookupBlock]],
    kerning_per_direction: dict[Direction, list[KerningPair]],
    ignoreMarks: bool = True,
    suffix: str = "",
) -> None:
    bidiGlyphs = context.bidiGlyphs
    side1Classes = context.kerning.side1Classes
    side2Classes = context.kerning.side2Classes

    newClassDefs, newSide1Classes, newSide2Classes = make_all_glyph_class_definitions(
        kerning_per_direction, context, context.feaFile
    )
    # NOTE: Consider duplicate names a bug, even if the classes would carry
    # the same glyphs.
    assert not context.kerning.classDefs.keys() & newClassDefs.keys()
    context.kerning.classDefs.update(newClassDefs)
    assert not side1Classes.keys() & newSide1Classes.keys()
    side1Classes.update(newSide1Classes)
    assert not side2Classes.keys() & newSide2Classes.keys()
    side2Classes.update(newSide2Classes)

    for direction, pairs in kerning_per_direction.items():
        lookupName = f"kern_{direction.value}{suffix}"
        lookup = make_kerning_lookup(
            context, options, lookupName, ignoreMarks=ignoreMarks
        )
        for pair in pairs:
            bidiTypes = {
                direction
                for direction, glyphs in bidiGlyphs.items()
                if not set(pair.glyphs).isdisjoint(glyphs)
            }
            if bidiTypes.issuperset(AMBIGUOUS_BIDIS):
                assert None, "this should have been caught by the splitter"
            # European and Arabic Numbers are always shaped LTR even in RTL scripts:
            pairIsRtl = (
                direction == Direction.RightToLeft
                and Direction.LeftToRight not in bidiTypes
            )
            rule = make_pairpos_rule(pair, side1Classes, side2Classes, pairIsRtl)
            lookup.statements.append(rule)
        lookups.setdefault(direction, {})[lookupName] = lookup


def make_all_glyph_class_definitions(
    kerning_per_direction: dict[Direction, list[KerningPair]],
    context: KernContext,
    feaFile: fea_ast.FeatureFile | None = None,
):
    # Note: Refer to the context for existing classDefs and mappings of glyph
    # class tuples to feaLib AST to avoid overwriting existing class names,
    # because base and mark kerning pairs might be separate passes.
    newClassDefs = {}
    existingSide1Classes = context.kerning.side1Classes
    existingSide2Classes = context.kerning.side2Classes
    newSide1Classes = {}
    newSide2Classes = {}
    side1Membership = context.side1Membership
    side2Membership = context.side2Membership

    if feaFile is not None:
        classNames = {cdef.name for cdef in ast.iterClassDefinitions(feaFile)}
    else:
        classNames = set()
    classNames.update(context.kerning.classDefs.keys())

    # Generate common class names first so that common classes are correctly
    # named in other lookups.
    for direction in (
        Direction.Neutral,
        Direction.LeftToRight,
        Direction.RightToLeft,
    ):
        for pair in kerning_per_direction.get(direction, []):
            if (
                pair.firstIsClass
                and pair.side1 not in existingSide1Classes
                and pair.side1 not in newSide1Classes
            ):
                addClassDefinition(
                    "kern1",
                    pair.side1,
                    newSide1Classes,
                    side1Membership,
                    newClassDefs,
                    classNames,
                    direction.value,
                )
            if (
                pair.secondIsClass
                and pair.side2 not in existingSide2Classes
                and pair.side2 not in newSide2Classes
            ):
                addClassDefinition(
                    "kern2",
                    pair.side2,
                    newSide2Classes,
                    side2Membership,
                    newClassDefs,
                    classNames,
                    direction.value,
                )

    return newClassDefs, newSide1Classes, newSide2Classes


def make_kerning_lookup(
    context: KernContext, options: SimpleNamespace, name: str, ignoreMarks: bool = True
) -> fea_ast.LookupBlock:
    lookup = fea_ast.LookupBlock(name)
    if ignoreMarks and options.ignoreMarks:
        # We only want to filter the spacing marks
        marks = set(context.gdefClasses.mark or []) & set(context.glyphSet.keys())

        spacing = []
        if marks:
            spacing = filter_spacing_marks(context, marks)
        if not spacing:
            # Simple case, there are no spacing ("Spacing Combining") marks,
            # do what we've always done.
            lookup.statements.append(ast.makeLookupFlag("IgnoreMarks"))
        else:
            # We want spacing marks to block kerns.
            className = f"MFS_{name}"
            filteringClass = ast.makeGlyphClassDefinitions(
                {className: spacing}, feaFile=context.feaFile
            )[className]
            lookup.statements.append(filteringClass)
            lookup.statements.append(
                ast.makeLookupFlag(markFilteringSet=filteringClass)
            )
    return lookup


def filter_spacing_marks(context: KernContext, marks: set[str]) -> list[str]:
    if context.isVariable:
        spacing = []
        for mark in marks:
            if all(
                source.font[mark].width != 0
                for source in context.font.sources
                if mark in source.font
            ):
                spacing.append(mark)
        return spacing

    return [mark for mark in marks if context.font[mark].width != 0]


def make_pairpos_rule(
    pair: KerningPair, side1Classes, side2Classes, rtl: bool = False
) -> fea_ast.PairPosStatement:
    enumerated = pair.firstIsClass ^ pair.secondIsClass
    valuerecord = fea_ast.ValueRecord(
        xPlacement=pair.value if rtl else None,
        yPlacement=0 if rtl else None,
        xAdvance=pair.value,
        yAdvance=0 if rtl else None,
    )

    if pair.firstIsClass:
        glyphs1 = fea_ast.GlyphClassName(side1Classes[pair.side1])
    else:
        glyphs1 = fea_ast.GlyphName(pair.side1)
    if pair.secondIsClass:
        glyphs2 = fea_ast.GlyphClassName(side2Classes[pair.side2])
    else:
        glyphs2 = fea_ast.GlyphName(pair.side2)

    return fea_ast.PairPosStatement(
        glyphs1=glyphs1,
        valuerecord1=valuerecord,
        glyphs2=glyphs2,
        valuerecord2=None,
        enumerated=enumerated,
    )


def make_feature_blocks(
    context: KernContext, lookups: dict[Direction, dict[str, Any]]
) -> Any:
    features = {}
    if "kern" in context.todo:
        kern = fea_ast.FeatureBlock("kern")
        register_lookups(context, kern, lookups)
        if kern.statements:
            features["kern"] = kern
    if "dist" in context.todo:
        dist = fea_ast.FeatureBlock("dist")
        register_lookups(context, dist, lookups)
        if dist.statements:
            features["dist"] = dist
    return features


def register_lookups(
    context: KernContext,
    feature: fea_ast.FeatureBlock,
    lookups: dict[Direction, dict[str, fea_ast.LookupBlock]],
) -> None:
    # Ensure we have kerning for pure common script runs (e.g. ">1")
    isKernBlock = feature.name == "kern"
    lookupsNeutral: list[fea_ast.LookupBlock] = []
    if isKernBlock and Direction.Neutral in lookups:
        lookupsNeutral.extend(
            lkp
            for lkp in lookups[Direction.Neutral].values()
            if lkp not in lookupsNeutral
        )

    # InDesign bugfix: register kerning lookups for all LTR scripts under DFLT
    # so that the basic composer, without a language selected, will still kern.
    # Register LTR lookups if any, otherwise RTL lookups.
    if isKernBlock:
        lookupsLTR: list[fea_ast.LookupBlock] = (
            list(lookups[Direction.LeftToRight].values())
            if Direction.LeftToRight in lookups
            else []
        )
        lookupsRTL: list[fea_ast.LookupBlock] = (
            list(lookups[Direction.RightToLeft].values())
            if Direction.RightToLeft in lookups
            else []
        )
        lookupsNeutral.extend(
            lkp for lkp in (lookupsLTR or lookupsRTL) if lkp not in lookupsNeutral
        )

    if lookupsNeutral:
        languages = context.feaLanguagesByTag.get("DFLT", ["dflt"])
        ast.addLookupReferences(feature, lookupsNeutral, "DFLT", languages)

    # Feature blocks use script tags to distinguish what to run for a
    # Unicode script.
    #
    # "Script tags generally correspond to a Unicode script. However, the
    # associations between them may not always be one-to-one, and the
    # OpenType script tags are not guaranteed to be the same as Unicode
    # Script property-value aliases or ISO 15924 script IDs."
    #
    # E.g. {"latn": "Latn", "telu": "Telu", "tel2": "Telu"}
    #
    # Skip DFLT script because we always take care of it above for `kern`.
    # It never occurs in `dist`.
    if isKernBlock:
        scriptsToReference: set[str] = context.knownScripts - DIST_ENABLED_SCRIPTS
    else:
        scriptsToReference = DIST_ENABLED_SCRIPTS.intersection(context.knownScripts)
    scriptsToReference -= DFLT_SCRIPTS
    for script in sorted(scriptsToReference):
        script_direction = script_horizontal_direction(script, "LTR")
        for tag in unicodedata.ot_tags_from_script(script):
            lookupsForThisScript = {}
            if Direction.Neutral in lookups:
                lookupsForThisScript.update(lookups[Direction.Neutral])
            if script_direction == "LTR" and Direction.LeftToRight in lookups:
                lookupsForThisScript.update(lookups[Direction.LeftToRight])
            if script_direction == "RTL" and Direction.RightToLeft in lookups:
                lookupsForThisScript.update(lookups[Direction.RightToLeft])
            if not lookupsForThisScript:
                continue
            if feature.statements:
                feature.statements.append(fea_ast.Comment(""))
            # Register the lookups for all languages defined in the feature
            # file for the script, otherwise kerning is not applied if any
            # language is set at all.
            languages = context.feaLanguagesByTag.get(tag, ["dflt"])
            ast.addLookupReferences(
                feature, lookupsForThisScript.values(), tag, languages
            )



import math

import pytest
from fontTools.designspaceLib import (
    AxisDescriptor,
    DesignSpaceDocument,
    SourceDescriptor,
)
from fontTools.misc.loggingTools import CapturingLogHandler
from fontTools.misc.transform import Transform

import ufo2ft.filters
from ufo2ft.constants import (
    _PRELIMINARY_CATEGORIES_KEY,
    GLYPHS_COMPONENT_INFO_KEY,
    OPENTYPE_CATEGORIES_KEY,
)
from ufo2ft.filters.propagateAnchors import (
    AnchorData,
    PropagateAnchorsFilter,
    PropagateAnchorsIFilter,
    _finalize_categories,
    get_xy_rotation,
    logger,
)
from ufo2ft.instantiator import Instantiator
from ufo2ft.util import _GlyphSet


@pytest.fixture(
    params=[
        {
            "glyphs": [
                {"name": "space", "width": 500},
                {
                    "name": "a",
                    "width": 350,
                    "outline": [
                        ("moveTo", ((0, 0),)),
                        ("lineTo", ((300, 0),)),
                        ("lineTo", ((300, 300),)),
                        ("lineTo", ((0, 300),)),
                        ("closePath", ()),
                    ],
                    "anchors": [(175, 300, "top"), (175, 0, "bottom")],
                },
                {
                    "name": "dieresiscomb",
                    "width": 0,
                    "outline": [
                        ("moveTo", ((-120, 320),)),
                        ("lineTo", ((-60, 320),)),
                        ("lineTo", ((-60, 360),)),
                        ("lineTo", ((-120, 360),)),
                        ("closePath", ()),
                        ("moveTo", ((120, 320),)),
                        ("lineTo", ((60, 320),)),
                        ("lineTo", ((60, 360),)),
                        ("lineTo", ((120, 360),)),
                        ("closePath", ()),
                    ],
                    "anchors": [(0, 300, "_top"), (0, 480, "top")],
                },
                {
                    "name": "macroncomb",
                    "width": 0,
                    "outline": [
                        ("moveTo", ((-120, 330),)),
                        ("lineTo", ((120, 330),)),
                        ("lineTo", ((120, 350),)),
                        ("lineTo", ((-120, 350),)),
                        ("closePath", ()),
                    ],
                    "anchors": [(0, 300, "_top"), (0, 480, "top")],
                },
                {
                    "name": "a-cyr",
                    "width": 350,
                    "outline": [("addComponent", ("a", (1, 0, 0, 1, 0, 0)))],
                },
                {
                    "name": "amacron",
                    "width": 350,
                    "outline": [
                        ("addComponent", ("a", (1, 0, 0, 1, 0, 0))),
                        ("addComponent", ("macroncomb", (1, 0, 0, 1, 175, 0))),
                    ],
                    "anchors": [(176, 481, "top")],
                },
                {
                    "name": "adieresis",
                    "width": 350,
                    "outline": [
                        ("addComponent", ("a", (1, 0, 0, 1, 0, 0))),
                        ("addComponent", ("dieresiscomb", (1, 0, 0, 1, 175, 0))),
                    ],
                },
                {
                    "name": "amacrondieresis",
                    "width": 350,
                    "outline": [
                        ("addComponent", ("amacron", (1, 0, 0, 1, 0, 0))),
                        ("addComponent", ("dieresiscomb", (1, 0, 0, 1, 175, 180))),
                    ],
                },
                {
                    "name": "adieresismacron",
                    "width": 350,
                    "outline": [
                        ("addComponent", ("a", (1, 0, 0, 1, 0, 0))),
                        ("addComponent", ("dieresiscomb", (1, 0, 0, 1, 175, 0))),
                        ("addComponent", ("macroncomb", (1, 0, 0, 1, 175, 180))),
                    ],
                },
                {
                    "name": "a_a",
                    "width": 700,
                    "outline": [
                        ("addComponent", ("a", (1, 0, 0, 1, 0, 0))),
                        ("addComponent", ("a", (1, 0, 0, 1, 350, 0))),
                    ],
                },
                {
                    "name": "emacron",
                    "width": 350,
                    "outline": [
                        ("addComponent", ("e", (1, 0, 0, 1, 0, 0))),
                        ("addComponent", ("macroncomb", (1, 0, 0, 1, 175, 0))),
                    ],
                },
                {
                    "name": "r",
                    "width": 350,
                    "outline": [
                        ("moveTo", ((0, 0),)),
                        ("lineTo", ((0, 300),)),
                        ("lineTo", ((175, 300),)),
                        ("closePath", ()),
                    ],
                    "anchors": [(175, 300, "top"), (175, 0, "bottom")],
                },
                {
                    "name": "rcombbelow",
                    "width": 0,
                    "outline": [
                        ("addComponent", ("r", (0.5, 0, 0, 0.5, -100, -100))),
                    ],
                    "anchors": [(0, 0, "_bottom")],
                },
            ]
        }
    ]
)
def font(request, FontClass):
    font = FontClass()
    for param in request.param["glyphs"]:
        glyph = font.newGlyph(param["name"])
        glyph.width = param.get("width", 0)
        pen = glyph.getPen()
        for operator, operands in param.get("outline", []):
            getattr(pen, operator)(*operands)
        for x, y, name in param.get("anchors", []):
            glyph.appendAnchor(dict(x=x, y=y, name=name))
    # classify as 'mark' all glyphs with zero width and 'comb' in their name
    categories = {
        g["name"]: "mark"
        for g in request.param["glyphs"]
        if g.get("width", 0) == 0 and "comb" in g["name"]
    }
    categories["a_a"] = "ligature"
    font.lib["public.openTypeCategories"] = categories
    return font


EXPECTED = {
    # single component glyph
    "a-cyr": ([("bottom", 175, 0), ("top", 175, 300)], {"a-cyr"}),
    # two component glyph
    "adieresis": ([("bottom", 175, 0), ("top", 175, 480)], {"adieresis"}),
    # one anchor, two component glyph (amacron has its own 'top' anchor)
    "amacron": ([("bottom", 175, 0), ("top", 176, 481)], {"amacron"}),
    # three component glyph
    "adieresismacron": ([("bottom", 175, 0), ("top", 175, 660)], {"adieresismacron"}),
    # nested component glyph
    "amacrondieresis": (
        [("bottom", 175, 0), ("top", 175, 660)],
        # 'amacron' is used as component by 'amacrondieresis' hence it is modified
        # as well...
        {"amacrondieresis", "amacron"},
    ),
    # ligature glyph
    "a_a": (
        [
            ("bottom_1", 175, 0),
            ("bottom_2", 525, 0),
            ("top_1", 175, 300),
            ("top_2", 525, 300),
        ],
        {"a_a"},
    ),
    # partial propagation: 'e' component is missing, but macroncomb's anchors
    # still propagate (with a warning about the missing component)
    "emacron": ([("top", 175, 480)], {"emacron"}),
    # the composite glyph is a mark with anchors, hence propagation is not performed,
    # i.e. 'top' and 'bottom' are *not* copied to 'rcombbelow':
    # https://github.com/googlefonts/ufo2ft/issues/802
    "rcombbelow": ([("_bottom", 0, 0)], set()),
}


class PropagateAnchorsFilterTest:
    def test_empty_glyph(self, font):
        philter = PropagateAnchorsFilter(include={"space"})
        assert not philter(font)

    def test_contour_glyph(self, font):
        philter = PropagateAnchorsFilter(include={"a"})
        assert not philter(font)

    @pytest.mark.parametrize("name", list(EXPECTED))
    def test_include_one_glyph_at_a_time(self, font, name):
        philter = PropagateAnchorsFilter(include={name})
        modified = philter(font)

        expected_anchors, expected_modified = EXPECTED[name]
        assert modified == expected_modified
        assert [(a.name, a.x, a.y) for a in font[name].anchors] == expected_anchors

    def test_whole_font(self, font):
        philter = PropagateAnchorsFilter()
        modified = philter(font)
        expected_modified = {k for k in EXPECTED if k in EXPECTED[k][1]}
        assert modified == expected_modified
        for name, (expected_anchors, _) in EXPECTED.items():
            assert [(a.name, a.x, a.y) for a in font[name].anchors] == expected_anchors

    def test_fail_during_anchor_propagation(self, font):
        name = "emacron"
        with CapturingLogHandler(logger, level="WARNING") as captor:
            philter = PropagateAnchorsFilter(include={name})
            philter(font)
        captor.assertRegex(
            "Anchors not propagated for inexistent component e in glyph emacron"
        )

    def test_logger(self, font):
        with CapturingLogHandler(logger, level="INFO") as captor:
            philter = PropagateAnchorsFilter()
            philter(font)
        captor.assertRegex("Glyphs with propagated anchors: 7")


def test_CantarellAnchorPropagation(FontClass, datadir):
    ufo_path = datadir.join("CantarellAnchorPropagation.ufo")
    ufo = FontClass(ufo_path)
    pre_filters, _ = ufo2ft.filters.loadFilters(ufo)

    philter = pre_filters[0]
    philter(ufo)

    anchors_combined = {
        (a.name, a.x, a.y) for a in ufo["circumflexcomb_tildecomb"].anchors
    }
    assert ("top", 214.0, 730.0) in anchors_combined
    assert ("_top", 213.0, 482.0) in anchors_combined

    anchors_o = {(a.name, a.x, a.y) for a in ufo["ocircumflextilde"].anchors}
    assert ("top", 284.0, 730.0) in anchors_o


def test_CantarellAnchorPropagation_reduced_filter(FontClass, datadir):
    ufo_path = datadir.join("CantarellAnchorPropagation.ufo")
    ufo = FontClass(ufo_path)
    ufo.lib["com.github.googlei18n.ufo2ft.filters"][0]["include"] = ["ocircumflextilde"]
    pre_filters, _ = ufo2ft.filters.loadFilters(ufo)

    philter = pre_filters[0]
    philter(ufo)

    anchors_combined = {
        (a.name, a.x, a.y) for a in ufo["circumflexcomb_tildecomb"].anchors
    }
    assert ("top", 214.0, 730.0) in anchors_combined
    assert ("_top", 213.0, 482.0) in anchors_combined

    anchors_o = {(a.name, a.x, a.y) for a in ufo["ocircumflextilde"].anchors}
    assert ("top", 284.0, 730.0) in anchors_o


# ---------------------------------------------------------------------------
# Tests ported from glyphsLib and fontc to keep the three implementations in sync.
# glyphsLib source: tests/builder/transformations/propagate_anchors_test.py
# fontc source: fontir/src/propagate_anchors.rs (test module)
#
# glyphsLib classifies marks/ligatures from GlyphData.xml (GSGlyph.category/
# subCategory); ufo2ft uses explicit public.openTypeCategories.  Where glyphsLib
# auto-detects a mark, we set categories explicitly here.
# ---------------------------------------------------------------------------


def _make_glyph(font, name, width=0, anchors=(), components=(), contour=True):
    """Helper matching glyphsLib's GlyphBuilder for defcon/ufoLib2 fonts."""
    g = font.newGlyph(name)
    g.width = width
    pen = g.getPen()
    if contour and not components:
        pen.moveTo((0, 0))
        pen.lineTo((width or 100, 0))
        pen.endPath()
    for base, pos in components:
        pen.addComponent(base, (1, 0, 0, 1, pos[0], pos[1]))
    for a in anchors:
        g.appendAnchor(dict(x=a[1][0], y=a[1][1], name=a[0]))
    return g


def _assert_anchors(font, glyph_name, expected):
    """Assert anchors match expected list of (name, (x, y)) tuples."""
    actual = [(a.name, (a.x, a.y)) for a in font[glyph_name].anchors]
    assert sorted(actual) == sorted(expected), (
        f"anchors for '{glyph_name}':\n"
        f"  actual:   {sorted(actual)}\n"
        f"  expected: {sorted(expected)}"
    )


def test_affine_scale():
    """Ported from glyphsLib test_affine_scale."""
    assert get_xy_rotation(
        Transform().translate(589, 502).rotate(math.radians(180))
    ) == (-1, -1)
    assert get_xy_rotation(Transform().translate(10, 10)) == (1, 1)
    assert get_xy_rotation(Transform().scale(1, -1)) == (1, -1)
    assert get_xy_rotation(Transform().scale(-1, 1)) == (-1, 1)
    assert get_xy_rotation(
        Transform().translate(589, 502).rotate(math.radians(180)).scale(-1, 1)
    ) == (1, -1)


class PortedAlgorithmTest:
    """Tests ported from glyphsLib/fontc to verify the three implementations
    stay in sync.  Uses the same glyph names and coordinates as the originals."""

    def test_no_components_anchors_are_unchanged(self, FontClass):
        """Ported from glyphsLib test_no_components_anchors_are_unchanged."""
        font = FontClass()
        _make_glyph(
            font,
            "A",
            600,
            anchors=[
                ("bottom", (234, 0)),
                ("ogonek", (411, 0)),
                ("top", (234, 810)),
            ],
        )
        _make_glyph(
            font,
            "acutecomb",
            0,
            anchors=[
                ("_top", (0, 578)),
                ("top", (0, 810)),
            ],
        )
        font.lib["public.openTypeCategories"] = {"acutecomb": "mark"}

        philter = PropagateAnchorsFilter()
        philter(font)

        _assert_anchors(
            font,
            "A",
            [
                ("bottom", (234, 0)),
                ("ogonek", (411, 0)),
                ("top", (234, 810)),
            ],
        )
        _assert_anchors(
            font,
            "acutecomb",
            [
                ("_top", (0, 578)),
                ("top", (0, 810)),
            ],
        )

    def test_basic_composite_anchor(self, FontClass):
        """Ported from glyphsLib test_basic_composite_anchor."""
        font = FontClass()
        _make_glyph(
            font,
            "A",
            600,
            anchors=[
                ("bottom", (234, 0)),
                ("ogonek", (411, 0)),
                ("top", (234, 810)),
            ],
        )
        _make_glyph(
            font,
            "acutecomb",
            0,
            anchors=[
                ("_top", (0, 578)),
                ("top", (0, 810)),
            ],
        )
        _make_glyph(
            font,
            "Aacute",
            600,
            contour=False,
            components=[
                ("A", (0, 0)),
                ("acutecomb", (234, 232)),
            ],
        )
        font.lib["public.openTypeCategories"] = {"acutecomb": "mark"}

        philter = PropagateAnchorsFilter()
        philter(font)

        _assert_anchors(
            font,
            "Aacute",
            [
                ("bottom", (234, 0)),
                ("ogonek", (411, 0)),
                ("top", (234, 1042)),
            ],
        )

    def test_propagate_ligature_anchors(self, FontClass):
        """Ported from glyphsLib test_propagate_ligature_anchors.
        Based on the IJ glyph in Oswald (ExtraLight)."""
        font = FontClass()
        _make_glyph(
            font,
            "I",
            206,
            anchors=[
                ("bottom", (103, 0)),
                ("ogonek", (103, 0)),
                ("top", (103, 810)),
                ("topleft", (20, 810)),
            ],
        )
        _make_glyph(
            font,
            "J",
            266,
            anchors=[
                ("bottom", (133, 0)),
                ("top", (163, 810)),
            ],
        )
        _make_glyph(
            font,
            "IJ",
            472,
            contour=False,
            components=[
                ("I", (0, 0)),
                ("J", (206, 0)),
            ],
        )
        font.lib["public.openTypeCategories"] = {"IJ": "ligature"}

        philter = PropagateAnchorsFilter()
        philter(font)

        _assert_anchors(
            font,
            "IJ",
            [
                ("bottom_1", (103, 0)),
                ("ogonek_1", (103, 0)),
                ("top_1", (103, 810)),
                ("topleft_1", (20, 810)),
                ("bottom_2", (339, 0)),
                ("top_2", (369, 810)),
            ],
        )

    def test_digraphs_arent_ligatures(self, FontClass):
        """Ported from glyphsLib test_digraphs_arent_ligatures.
        Same glyphs as test_propagate_ligature_anchors but IJ is NOT
        classified as ligature — anchors are not numbered."""
        font = FontClass()
        _make_glyph(
            font,
            "I",
            206,
            anchors=[
                ("bottom", (103, 0)),
                ("ogonek", (103, 0)),
                ("top", (103, 810)),
                ("topleft", (20, 810)),
            ],
        )
        _make_glyph(
            font,
            "J",
            266,
            anchors=[
                ("bottom", (133, 0)),
                ("top", (163, 810)),
            ],
        )
        _make_glyph(
            font,
            "IJ",
            472,
            contour=False,
            components=[
                ("I", (0, 0)),
                ("J", (206, 0)),
            ],
        )
        font.lib["public.openTypeCategories"] = {}

        philter = PropagateAnchorsFilter()
        philter(font)

        _assert_anchors(
            font,
            "IJ",
            [
                ("bottom", (339, 0)),
                ("ogonek", (103, 0)),
                ("top", (369, 810)),
                ("topleft", (20, 810)),
            ],
        )

    def test_remove_exit_anchor_on_component(self, FontClass):
        """Ported from glyphsLib test_remove_exit_anchor_on_component."""
        font = FontClass()
        _make_glyph(font, "comma", 250)
        _make_glyph(
            font,
            "ain-ar.init",
            400,
            anchors=[
                ("top", (294, 514)),
                ("exit", (0, 0)),
            ],
        )
        _make_glyph(
            font,
            "ain-ar.init.alt",
            400,
            contour=False,
            components=[
                ("ain-ar.init", (0, 0)),
                ("comma", (0, 0)),
            ],
        )
        font.lib["public.openTypeCategories"] = {}

        philter = PropagateAnchorsFilter()
        philter(font)

        _assert_anchors(font, "ain-ar.init.alt", [("top", (294, 514))])

    def test_component_anchor(self, FontClass):
        """Ported from glyphsLib test_component_anchor / fontc component_anchor."""
        font = FontClass()
        _make_glyph(
            font,
            "acutecomb",
            0,
            anchors=[
                ("_top", (150, 580)),
                ("top", (170, 792)),
            ],
        )
        g = _make_glyph(
            font,
            "aa",
            960,
            anchors=[
                ("bottom_1", (218, 8)),
                ("bottom_2", (742, 7)),
                ("ogonek_1", (398, 9)),
                ("ogonek_2", (902, 9)),
                ("top_1", (227, 548)),
                ("top_2", (746, 548)),
            ],
        )
        _make_glyph(
            font,
            "a_a",
            960,
            contour=False,
            components=[
                ("aa", (0, 0)),
            ],
        )
        g = _make_glyph(
            font,
            "a_aacute",
            960,
            contour=False,
            components=[
                ("a_a", (0, 0)),
                ("acutecomb", (596, -32)),
            ],
        )
        # UFO components have no lib, so glyphsLib stores per-component
        # properties in the composite glyph's lib; index 1 = acutecomb
        g.lib[GLYPHS_COMPONENT_INFO_KEY] = [
            {"index": 1, "anchor": "top_2"},
        ]
        font.lib["public.openTypeCategories"] = {
            "acutecomb": "mark",
            "aa": "ligature",
            "a_a": "ligature",
            "a_aacute": "ligature",
        }

        philter = PropagateAnchorsFilter()
        philter(font)

        _assert_anchors(
            font,
            "a_aacute",
            [
                ("bottom_1", (218, 8)),
                ("bottom_2", (742, 7)),
                ("ogonek_1", (398, 9)),
                ("ogonek_2", (902, 9)),
                ("top_1", (227, 548)),
                # top_2 replaced by acutecomb's "top" renamed via ComponentInfo:
                # (170, 792) + offset (596, -32) = (766, 760)
                ("top_2", (766, 760)),
            ],
        )

    def test_origin_anchor(self, FontClass):
        """Ported from glyphsLib test_origin_anchor."""
        font = FontClass()
        _make_glyph(
            font,
            "a",
            500,
            anchors=[
                ("*origin", (-20, 0)),
                ("bottom", (242, 7)),
                ("ogonek", (402, 9)),
                ("top", (246, 548)),
            ],
        )
        _make_glyph(
            font,
            "acutecomb",
            0,
            anchors=[
                ("_top", (150, 580)),
                ("top", (170, 792)),
            ],
        )
        _make_glyph(
            font,
            "aacute",
            500,
            contour=False,
            components=[
                ("a", (0, 0)),
                ("acutecomb", (116, -32)),
            ],
        )
        font.lib["public.openTypeCategories"] = {"acutecomb": "mark"}

        philter = PropagateAnchorsFilter()
        philter(font)

        _assert_anchors(
            font,
            "aacute",
            [
                ("bottom", (262, 7)),
                ("ogonek", (422, 9)),
                ("top", (286, 760)),
            ],
        )

    def test_invert_names_on_rotation(self, FontClass):
        """Ported from glyphsLib test_invert_names_on_rotation."""
        font = FontClass()
        _make_glyph(font, "comma", 250)
        _make_glyph(
            font,
            "commaaccentcomb",
            0,
            anchors=[
                ("_bottom", (289, 0)),
                ("mybottom", (277, -308)),
            ],
        )
        # Add a component to commaaccentcomb
        pen = font["commaaccentcomb"].getPen()
        pen.addComponent("comma", (1, 0, 0, 1, 9, -164))

        # 180° rotated component
        g = font.newGlyph("commaturnedabovecomb")
        g.width = 0
        pen = g.getPen()
        t = Transform().translate(589, 502).rotate(math.radians(180))
        pen.addComponent("commaaccentcomb", tuple(t))
        font.lib["public.openTypeCategories"] = {
            "commaaccentcomb": "mark",
            "commaturnedabovecomb": "mark",
        }

        philter = PropagateAnchorsFilter()
        philter(font)

        _assert_anchors(
            font,
            "commaturnedabovecomb",
            [
                ("_top", (300, 502)),
                ("mytop", (312, 810)),
            ],
        )

    def test_entry_anchor_on_non_first_component(self, FontClass):
        """Ported from glyphsLib test_entry_anchor_on_non_first_component."""
        font = FontClass()
        _make_glyph(font, "part1", 200, anchors=[("top", (10, 0))])
        _make_glyph(font, "part2", 200, anchors=[("entry.2", (10, 0))])
        _make_glyph(
            font,
            "combo",
            400,
            contour=False,
            components=[
                ("part1", (0, 0)),
                ("part2", (100, 0)),
            ],
        )
        font.lib["public.openTypeCategories"] = {}

        philter = PropagateAnchorsFilter()
        philter(font)

        _assert_anchors(font, "combo", [("top", (10, 0))])

    def test_cursive_anchors_ligature(self, FontClass):
        """Ported from glyphsLib test_cursive_anchors_ligature."""
        font = FontClass()
        _make_glyph(
            font,
            "part1_part2",
            200,
            anchors=[
                ("entry.1", (10, 0)),
                ("exit.1", (100, 0)),
            ],
        )
        _make_glyph(
            font,
            "combo",
            200,
            contour=False,
            components=[
                ("part1_part2", (0, 0)),
            ],
        )
        font.lib["public.openTypeCategories"] = {"combo": "ligature"}

        philter = PropagateAnchorsFilter()
        philter(font)

        _assert_anchors(
            font,
            "combo",
            [
                ("entry.1", (10, 0)),
                ("exit.1", (100, 0)),
            ],
        )


class CharacterizationTest:
    """Tests for ufo2ft-specific behavior not covered by the ported suite."""

    def test_mixed_mark_base_anchor_component(self, FontClass):
        """A component with both _bottomright (mark-attaching) and bottom
        (base anchor for mark-to-mark stacking) should propagate the base
        anchor. The old algorithm classifies the entire component as 'mark'
        and drops all its anchors. This documents the known bug."""
        font = FontClass()
        # Base glyph with a 'top' anchor
        g = font.newGlyph("A")
        g.width = 600
        pen = g.getPen()
        pen.moveTo((0, 0))
        pen.lineTo((600, 0))
        pen.lineTo((300, 700))
        pen.closePath()
        g.appendAnchor(dict(x=300, y=700, name="top"))
        g.appendAnchor(dict(x=300, y=0, name="bottom"))

        # Mark component with both _bottomright (mark-attaching) and
        # bottom (base anchor for stacking another mark below)
        g = font.newGlyph("ogonek")
        g.width = 0
        pen = g.getPen()
        pen.moveTo((0, 0))
        pen.lineTo((100, -200))
        pen.lineTo((50, -100))
        pen.closePath()
        g.appendAnchor(dict(x=50, y=0, name="_bottomright"))
        g.appendAnchor(dict(x=50, y=-200, name="bottom"))

        # Composite: A + ogonek
        g = font.newGlyph("Aogonek")
        g.width = 600
        pen = g.getPen()
        pen.addComponent("A", (1, 0, 0, 1, 0, 0))
        pen.addComponent("ogonek", (1, 0, 0, 1, 300, 0))

        font.lib["public.openTypeCategories"] = {"ogonek": "mark"}

        philter = PropagateAnchorsFilter()
        philter(font)

        anchors = {(a.name, a.x, a.y) for a in font["Aogonek"].anchors}

        # The new algorithm processes each anchor individually, so ogonek's
        # "bottom" anchor (a base anchor for mark-to-mark stacking) is
        # correctly propagated. ogonek's "bottom" overwrites A's "bottom"
        # because ogonek attaches to A and the mark's stacking anchor
        # takes precedence.
        assert ("top", 300, 700) in anchors
        assert ("bottom", 350, -200) in anchors

    def test_mark_with_existing_anchors_skips_propagation(self, FontClass):
        """A mark glyph that already has anchors should not have component
        anchors propagated into it."""
        font = FontClass()
        g = font.newGlyph("dotbelow")
        g.width = 0
        pen = g.getPen()
        pen.moveTo((-50, -100))
        pen.lineTo((50, -100))
        pen.lineTo((50, 0))
        pen.lineTo((-50, 0))
        pen.closePath()
        g.appendAnchor(dict(x=0, y=-100, name="_bottom"))
        g.appendAnchor(dict(x=0, y=0, name="bottom"))

        g = font.newGlyph("brevecomb")
        g.width = 0
        pen = g.getPen()
        pen.moveTo((-60, 500))
        pen.lineTo((60, 500))
        pen.lineTo((0, 550))
        pen.closePath()
        g.appendAnchor(dict(x=0, y=480, name="_top"))
        g.appendAnchor(dict(x=0, y=580, name="top"))

        # A mark composed of another mark, with its own anchors
        g = font.newGlyph("dotbelow_special")
        g.width = 0
        pen = g.getPen()
        pen.addComponent("dotbelow", (1, 0, 0, 1, 0, 0))
        g.appendAnchor(dict(x=10, y=-110, name="_bottom"))
        g.appendAnchor(dict(x=10, y=10, name="bottom"))

        font.lib["public.openTypeCategories"] = {
            "dotbelow": "mark",
            "brevecomb": "mark",
            "dotbelow_special": "mark",
        }

        philter = PropagateAnchorsFilter()
        modified = philter(font)

        # Mark with own anchors: propagation skipped
        assert "dotbelow_special" not in modified
        anchors = [(a.name, a.x, a.y) for a in font["dotbelow_special"].anchors]
        assert anchors == [("_bottom", 10, -110), ("bottom", 10, 10)]

    def test_nested_composites_shared_component(self, FontClass):
        """Nested composites sharing a common component should get consistent
        anchor propagation regardless of processing order."""
        font = FontClass()
        g = font.newGlyph("base")
        g.width = 500
        pen = g.getPen()
        pen.moveTo((0, 0))
        pen.lineTo((500, 0))
        pen.lineTo((250, 600))
        pen.closePath()
        g.appendAnchor(dict(x=250, y=600, name="top"))
        g.appendAnchor(dict(x=250, y=0, name="bottom"))

        g = font.newGlyph("mark1")
        g.width = 0
        pen = g.getPen()
        pen.moveTo((-30, 0))
        pen.lineTo((30, 0))
        pen.lineTo((0, 50))
        pen.closePath()
        g.appendAnchor(dict(x=0, y=0, name="_top"))
        g.appendAnchor(dict(x=0, y=50, name="top"))

        g = font.newGlyph("mark2")
        g.width = 0
        pen = g.getPen()
        pen.moveTo((-40, 0))
        pen.lineTo((40, 0))
        pen.lineTo((0, 60))
        pen.closePath()
        g.appendAnchor(dict(x=0, y=0, name="_top"))
        g.appendAnchor(dict(x=0, y=60, name="top"))

        # shared_comp uses base
        g = font.newGlyph("shared_comp")
        g.width = 500
        pen = g.getPen()
        pen.addComponent("base", (1, 0, 0, 1, 0, 0))

        # comp_a = shared_comp + mark1
        g = font.newGlyph("comp_a")
        g.width = 500
        pen = g.getPen()
        pen.addComponent("shared_comp", (1, 0, 0, 1, 0, 0))
        pen.addComponent("mark1", (1, 0, 0, 1, 250, 0))

        # comp_b = shared_comp + mark2
        g = font.newGlyph("comp_b")
        g.width = 500
        pen = g.getPen()
        pen.addComponent("shared_comp", (1, 0, 0, 1, 0, 0))
        pen.addComponent("mark2", (1, 0, 0, 1, 250, 0))

        font.lib["public.openTypeCategories"] = {
            "mark1": "mark",
            "mark2": "mark",
        }

        philter = PropagateAnchorsFilter()
        philter(font)

        # shared_comp should have same anchors as base
        assert {(a.name, a.x, a.y) for a in font["shared_comp"].anchors} == {
            ("top", 250, 600),
            ("bottom", 250, 0),
        }

        # comp_a: top replaced by mark1's top position (transformed)
        anchors_a = {(a.name, a.x, a.y) for a in font["comp_a"].anchors}
        assert ("bottom", 250, 0) in anchors_a
        assert ("top", 250, 50) in anchors_a  # mark1's top at (0+250, 50+0)

        # comp_b: top replaced by mark2's top position (transformed)
        anchors_b = {(a.name, a.x, a.y) for a in font["comp_b"].anchors}
        assert ("bottom", 250, 0) in anchors_b
        assert ("top", 250, 60) in anchors_b  # mark2's top at (0+250, 60+0)

    def test_anchor_ordering_deterministic(self, FontClass):
        """Propagated anchors should be in a deterministic order."""
        font = FontClass()
        g = font.newGlyph("base")
        g.width = 500
        pen = g.getPen()
        pen.moveTo((0, 0))
        pen.lineTo((500, 0))
        pen.lineTo((250, 600))
        pen.closePath()
        g.appendAnchor(dict(x=250, y=600, name="top"))
        g.appendAnchor(dict(x=250, y=0, name="bottom"))
        g.appendAnchor(dict(x=250, y=300, name="center"))

        g = font.newGlyph("comp")
        g.width = 500
        pen = g.getPen()
        pen.addComponent("base", (1, 0, 0, 1, 0, 0))

        font.lib["public.openTypeCategories"] = {}

        philter = PropagateAnchorsFilter()
        philter(font)

        names = [a.name for a in font["comp"].anchors]
        assert names == sorted(names)

    def test_include_root_with_transitive_deps(self, FontClass):
        """When include is set, transitive component dependencies must also
        be written (the reduced Cantarell test depends on this)."""
        font = FontClass()
        g = font.newGlyph("base")
        g.width = 500
        pen = g.getPen()
        pen.moveTo((0, 0))
        pen.lineTo((500, 0))
        pen.lineTo((250, 600))
        pen.closePath()
        g.appendAnchor(dict(x=250, y=600, name="top"))

        g = font.newGlyph("mid")
        g.width = 500
        pen = g.getPen()
        pen.addComponent("base", (1, 0, 0, 1, 0, 0))

        g = font.newGlyph("topglyph")
        g.width = 500
        pen = g.getPen()
        pen.addComponent("mid", (1, 0, 0, 1, 0, 0))

        font.lib["public.openTypeCategories"] = {}

        # Only include the top-level glyph
        philter = PropagateAnchorsFilter(include={"topglyph"})
        modified = philter(font)

        # Both mid and topglyph should be modified (mid is a transitive dep)
        assert "mid" in modified
        assert "topglyph" in modified
        assert [(a.name, a.x, a.y) for a in font["mid"].anchors] == [("top", 250, 600)]
        assert [(a.name, a.x, a.y) for a in font["topglyph"].anchors] == [
            ("top", 250, 600)
        ]


class AlgorithmDetailTest:
    """Tests for ufo2ft-specific algorithm details not covered by the ported suite."""

    def test_bottom_top_cancellation(self, FontClass):
        """When a non-mark composite has _bottom in its own anchors,
        propagated 'top' and '_top' from components are removed."""
        font = FontClass()
        g = font.newGlyph("base")
        g.width = 500
        pen = g.getPen()
        pen.moveTo((0, 0))
        pen.lineTo((500, 0))
        pen.lineTo((250, 600))
        pen.closePath()
        g.appendAnchor(dict(x=250, y=600, name="top"))
        g.appendAnchor(dict(x=250, y=0, name="bottom"))

        # Non-mark composite that declares _bottom (attaches below something)
        # but is NOT classified as mark, so propagation runs
        g = font.newGlyph("comp")
        g.width = 500
        pen = g.getPen()
        pen.addComponent("base", (1, 0, 0, 1, 0, 0))
        g.appendAnchor(dict(x=250, y=-50, name="_bottom"))

        # Explicitly classify as base so the mark heuristic doesn't trigger
        font.lib["public.openTypeCategories"] = {"comp": "base"}

        philter = PropagateAnchorsFilter()
        philter(font)

        anchors = {(a.name, a.x, a.y) for a in font["comp"].anchors}
        # _bottom is present on the composite itself, so propagated 'top'
        # and '_top' from components should be cancelled
        assert ("_bottom", 250, -50) in anchors
        assert ("bottom", 250, 0) in anchors
        assert not any(a[0] == "top" for a in anchors)
        assert not any(a[0] == "_top" for a in anchors)

    def test_absent_categories_fallback_heuristics(self, FontClass):
        """When public.openTypeCategories is completely absent, the algorithm
        falls back to heuristics: _-prefixed anchors -> mark, _ in name -> ligature."""
        font = FontClass()
        g = font.newGlyph("a")
        g.width = 350
        pen = g.getPen()
        pen.moveTo((0, 0))
        pen.lineTo((300, 0))
        pen.lineTo((300, 300))
        pen.lineTo((0, 300))
        pen.closePath()
        g.appendAnchor(dict(x=150, y=300, name="top"))
        g.appendAnchor(dict(x=150, y=0, name="bottom"))

        g = font.newGlyph("acutecomb")
        g.width = 0
        pen = g.getPen()
        pen.moveTo((0, 500))
        pen.lineTo((50, 600))
        pen.lineTo((-50, 600))
        pen.closePath()
        g.appendAnchor(dict(x=0, y=500, name="_top"))
        g.appendAnchor(dict(x=0, y=620, name="top"))

        # Mark (inferred from _top anchor)
        g = font.newGlyph("aacute")
        g.width = 350
        pen = g.getPen()
        pen.addComponent("a", (1, 0, 0, 1, 0, 0))
        pen.addComponent("acutecomb", (1, 0, 0, 1, 150, 0))

        # Ligature (inferred from _ in name)
        g = font.newGlyph("a_a")
        g.width = 700
        pen = g.getPen()
        pen.addComponent("a", (1, 0, 0, 1, 0, 0))
        pen.addComponent("a", (1, 0, 0, 1, 350, 0))

        # No categories at all
        font.lib["public.openTypeCategories"] = {}

        from fontTools.misc.loggingTools import CapturingLogHandler

        with CapturingLogHandler(logger, level="WARNING") as captor:
            philter = PropagateAnchorsFilter()
            philter(font)

        captor.assertRegex("public.openTypeCategories not found or empty")

        # aacute: mark attachment works via heuristic
        anchors_aacute = {(a.name, a.x, a.y) for a in font["aacute"].anchors}
        assert ("bottom", 150, 0) in anchors_aacute
        assert ("top", 150, 620) in anchors_aacute

        # a_a: no ligature inference from name alone, and no caret_N anchors;
        # second component's anchors overwrite first's
        anchors_aa = {(a.name, a.x, a.y) for a in font["a_a"].anchors}
        assert ("top", 500, 300) in anchors_aa
        assert ("bottom", 500, 0) in anchors_aa
        assert not any("_1" in a[0] or "_2" in a[0] for a in anchors_aa)

    def test_partial_categories(self, FontClass):
        """When public.openTypeCategories is present but partial, the fallback
        heuristic does NOT activate — only explicitly classified glyphs are
        marked/ligature."""
        font = FontClass()
        g = font.newGlyph("a")
        g.width = 350
        pen = g.getPen()
        pen.moveTo((0, 0))
        pen.lineTo((300, 0))
        pen.lineTo((300, 300))
        pen.lineTo((0, 300))
        pen.closePath()
        g.appendAnchor(dict(x=150, y=300, name="top"))
        g.appendAnchor(dict(x=150, y=0, name="bottom"))

        # a_a is NOT classified as ligature in the partial categories
        g = font.newGlyph("a_a")
        g.width = 700
        pen = g.getPen()
        pen.addComponent("a", (1, 0, 0, 1, 0, 0))
        pen.addComponent("a", (1, 0, 0, 1, 350, 0))

        # Partial categories: only 'a' is classified
        font.lib["public.openTypeCategories"] = {"a": "base"}

        philter = PropagateAnchorsFilter()
        philter(font)

        # Without ligature classification, second 'a' overwrites first
        anchors = {(a.name, a.x, a.y) for a in font["a_a"].anchors}
        assert ("top", 500, 300) in anchors
        assert ("bottom", 500, 0) in anchors
        assert not any("_1" in a[0] or "_2" in a[0] for a in anchors)


class PropagateAnchorsIFilterTest:
    def test_propagate_from_interpolated_components(self, FontClass, data_dir):
        ds_path = data_dir / "SkipExportGlyphsTest.designspace"
        ds = DesignSpaceDocument.fromfile(ds_path)
        ds.loadSourceFonts(FontClass)

        ufos = [s.font for s in ds.sources]
        glyphSets = [_GlyphSet.from_layer(s.font, s.layerName) for s in ds.sources]

        assert len(ufos) == len(glyphSets) == 4

        # the composite glyph 'Astroke' has no anchors, but 'A' has some
        for glyphSet in glyphSets:
            if "Astroke" in glyphSet:
                assert not glyphSet["Astroke"].anchors
            if "A" in glyphSet:
                assert glyphSet["A"].anchors

        # in glyphSets[2] the 'Astroke' component base glyphs are missing so their
        # propagated anchors are supposed to be interpolated on the fly
        assert "Astroke" in glyphSets[2]
        assert {c.baseGlyph for c in glyphSets[2]["Astroke"].components}.isdisjoint(
            glyphSets[2].keys()
        )
        assert not glyphSets[2]["Astroke"].anchors

        instantiator = Instantiator.from_designspace(
            ds, do_kerning=False, do_info=False
        )

        philter = PropagateAnchorsIFilter()

        modified = philter(ufos, glyphSets, instantiator)

        assert modified == {"Astroke"}

        assert [dict(a) for a in glyphSets[2]["Astroke"].anchors] == [
            {"name": "bottom", "x": 458, "y": 0},
            {"name": "center", "x": 457, "y": 358},
            {"name": "top", "x": 457, "y": 714},
            {"name": "topright", "x": 716, "y": 714},
        ]
        assert {c.baseGlyph for c in glyphSets[2]["Astroke"].components}.isdisjoint(
            glyphSets[2].keys()
        )

    def test_propagate_from_shared_nested_component_is_order_independent(
        self, FontClass, data_dir
    ):
        """Reproduces https://github.com/googlefonts/fontc/issues/1646

        This particular test font contains the following composite glyphs:
        - A-cy: uses A as base
        - Abreve-cy: uses A-cy and brevecomb as bases
        - Adieresis-cy: uses A-cy and dieresiscomb as bases
        The base glyph A has the following anchors: bottom, ogonek, top; A-cy has only
        top anchor; Abreve-cy and Adieresis-cy have no anchors.

        After PropagateAnchorsIFilterTest, the anchors of Abreve-cy and Adieresis-cy
        should be the same as A-cy.

        PropagateAnchorsIFilter processes glyphs with the same component depth in a
        non-deterministic order which varies because of hash randomization, which is ok
        because the algorithm is supposed to be order independent.
        However this exposed another bug whereby, depending on whether Abreve-cy or
        Adieresis-cy was processed first, only _one_ of the two glyphs would receive the
        anchors from A while the other would only get a single "top" anchor from A-cy.
        """
        ds_path = data_dir / "PropagateAnchorsIFilterTest.designspace"
        ds = DesignSpaceDocument.fromfile(ds_path)
        ds.loadSourceFonts(FontClass)

        ufos = [s.font for s in ds.sources]
        glyphSets = [_GlyphSet.from_layer(s.font, s.layerName) for s in ds.sources]

        assert len(ufos) == len(glyphSets) == 2

        expected_anchors = {"bottom", "ogonek", "top"}

        for glyphSet in glyphSets:
            # A has all three anchors
            assert {a.name for a in glyphSet["A"].anchors} == expected_anchors
            # A-cy has only top anchor and uses A as component
            assert {a.name for a in glyphSet["A-cy"].anchors} == {"top"}
            assert {c.baseGlyph for c in glyphSet["A-cy"].components} == {"A"}
            # Abreve-cy and Adieresis-cy both have no anchors and use A-cy as component
            assert not {a.name for a in glyphSet["Abreve-cy"].anchors}
            assert {c.baseGlyph for c in glyphSet["Abreve-cy"].components} == {
                "A-cy",
                "brevecomb.cap",
            }

            assert not {a.name for a in glyphSet["Adieresis-cy"].anchors}
            assert {c.baseGlyph for c in glyphSet["Adieresis-cy"].components} == {
                "A-cy",
                "dieresiscomb.cap",
            }

        instantiator = Instantiator.from_designspace(
            ds, do_kerning=False, do_info=False
        )

        philter = PropagateAnchorsIFilter()
        modified = philter(ufos, glyphSets, instantiator)

        # We expect all three glyphs to have the same anchor names after propagation
        assert modified == {"A-cy", "Abreve-cy", "Adieresis-cy"}
        for glyphSet in glyphSets:
            # but sometimes the first assert would fail, sometimes the second...
            assert {
                a.name for a in glyphSet["Abreve-cy"].anchors
            } == expected_anchors, "Abreve-cy"
            assert {
                a.name for a in glyphSet["Adieresis-cy"].anchors
            } == expected_anchors, "Adieresis-cy"
            assert {a.name for a in glyphSet["A-cy"].anchors} == expected_anchors

    def test_openTypeCategories_from_kwargs(self, FontClass):
        """When openTypeCategories is passed via kwargs (from the DS lib),
        the IFilter uses it for classification instead of reading from
        individual UFO libs."""
        font = FontClass()
        for name, width, anchors in [
            ("a", 350, [("top", 175, 300), ("bottom", 175, 0)]),
            ("b", 350, [("top", 175, 300), ("bottom", 175, 0)]),
        ]:
            g = font.newGlyph(name)
            g.width = width
            pen = g.getPen()
            pen.moveTo((0, 0))
            pen.lineTo((300, 0))
            pen.lineTo((300, 300))
            pen.lineTo((0, 300))
            pen.closePath()
            for aname, ax, ay in anchors:
                g.appendAnchor(dict(name=aname, x=ax, y=ay))
        # a_b is a composite ligature with no own anchors
        g = font.newGlyph("a_b")
        g.width = 700
        pen = g.getPen()
        pen.addComponent("a", (1, 0, 0, 1, 0, 0))
        pen.addComponent("b", (1, 0, 0, 1, 350, 0))

        # No categories in font.lib — the IFilter would fall back to heuristics
        assert "public.openTypeCategories" not in font.lib

        glyphSets = [_GlyphSet.from_layer(font)]
        philter = PropagateAnchorsIFilter()

        # Without categories: a_b gets unnumbered anchors (not a ligature)
        philter([font], glyphSets)
        anchors_no_cats = {a.name for a in glyphSets[0]["a_b"].anchors}
        assert "top_1" not in anchors_no_cats

        # Reset
        glyphSets = [_GlyphSet.from_layer(font)]

        # With DS-level categories via kwargs: a_b gets numbered anchors
        ds_categories = {"a_b": "ligature"}
        philter([font], glyphSets, openTypeCategories=ds_categories)
        anchor_names = {a.name for a in glyphSets[0]["a_b"].anchors}
        assert "top_1" in anchor_names
        assert "top_2" in anchor_names
        assert "bottom_1" in anchor_names
        assert "bottom_2" in anchor_names

    def test_preliminary_categories_finalization(self, FontClass):
        """preliminaryOpenTypeCategories are used for classification during
        propagation, then finalized (anchorless ligatures pruned, bases
        inferred) and written to font.lib."""
        font = FontClass()
        # Base glyph with attaching anchors
        g = font.newGlyph("a")
        g.width = 350
        pen = g.getPen()
        pen.moveTo((0, 0))
        pen.lineTo((300, 0))
        pen.lineTo((300, 300))
        pen.lineTo((0, 300))
        pen.closePath()
        g.appendAnchor(dict(name="top", x=150, y=300))
        g.appendAnchor(dict(name="bottom", x=150, y=0))

        # Mark glyph
        g = font.newGlyph("acutecomb")
        g.width = 0
        pen = g.getPen()
        pen.moveTo((0, 400))
        pen.lineTo((100, 500))
        pen.lineTo((0, 500))
        pen.closePath()
        g.appendAnchor(dict(name="_top", x=50, y=400))
        g.appendAnchor(dict(name="top", x=50, y=550))

        # Composite ligature a_a — should get numbered anchors
        g = font.newGlyph("a_a")
        g.width = 700
        pen = g.getPen()
        pen.addComponent("a", (1, 0, 0, 1, 0, 0))
        pen.addComponent("a", (1, 0, 0, 1, 350, 0))

        # Composite glyph, not explicitly categorized — should be inferred as base
        g = font.newGlyph("aacute")
        g.width = 350
        pen = g.getPen()
        pen.addComponent("a", (1, 0, 0, 1, 0, 0))
        pen.addComponent("acutecomb", (1, 0, 0, 1, 0, 0))

        # Ligature without any anchors — should be pruned
        g = font.newGlyph("noanchors_liga")
        g.width = 350
        pen = g.getPen()
        pen.moveTo((0, 0))
        pen.lineTo((300, 0))
        pen.lineTo((300, 300))
        pen.lineTo((0, 300))
        pen.closePath()

        # Spacing glyph (acute, not a combining mark) — should NOT become mark
        g = font.newGlyph("acute")
        g.width = 200
        pen = g.getPen()
        pen.addComponent("acutecomb", (1, 0, 0, 1, 100, 0))

        assert "public.openTypeCategories" not in font.lib

        preliminary = {
            "acutecomb": "mark",
            "a_a": "ligature",
            "noanchors_liga": "ligature",
        }

        glyphSets = [_GlyphSet.from_layer(font)]
        philter = PropagateAnchorsIFilter()
        philter([font], glyphSets, preliminaryOpenTypeCategories=preliminary)

        # Finalized categories should be written to font.lib
        assert "public.openTypeCategories" in font.lib
        cats = font.lib["public.openTypeCategories"]

        # Mark stays mark
        assert cats["acutecomb"] == "mark"
        # Ligature with attaching anchors stays ligature
        assert cats["a_a"] == "ligature"
        # Ligature without attaching anchors is pruned (not in result)
        assert "noanchors_liga" not in cats
        # Base glyph inferred from attaching anchors
        assert cats["a"] == "base"
        # Composite base inferred
        assert cats["aacute"] == "base"
        # Spacing acute: has propagated _top + top from acutecomb.
        # top is attaching, so it becomes base (not mark — not in preliminary marks)
        assert cats["acute"] == "base"

        # Check a_a got numbered anchors (ligature propagation worked)
        anchor_names = {a.name for a in glyphSets[0]["a_a"].anchors}
        assert "top_1" in anchor_names
        assert "top_2" in anchor_names

    def test_preliminary_categories_ignored_when_source_has_categories(self, FontClass):
        """Source-level openTypeCategories takes precedence over preliminary."""
        font = FontClass()
        g = font.newGlyph("a")
        g.width = 350
        pen = g.getPen()
        pen.moveTo((0, 0))
        pen.lineTo((300, 0))
        pen.lineTo((300, 300))
        pen.lineTo((0, 300))
        pen.closePath()
        g.appendAnchor(dict(name="top", x=150, y=300))

        # Source has explicit categories
        ds_categories = {"a": "base"}
        preliminary = {"a": "mark"}  # conflicting

        glyphSets = [_GlyphSet.from_layer(font)]
        philter = PropagateAnchorsIFilter()
        philter(
            [font],
            glyphSets,
            openTypeCategories=ds_categories,
            preliminaryOpenTypeCategories=preliminary,
        )

        # Source-level takes precedence: no finalized categories written
        # (source already has them)
        assert OPENTYPE_CATEGORIES_KEY not in font.lib

    def test_preliminary_categories_ignored_when_ufo_has_categories(self, FontClass):
        """UFO-level openTypeCategories takes precedence over preliminary."""
        font = FontClass()
        _make_glyph(
            font,
            "a",
            350,
            anchors=[("top", (150, 300)), ("bottom", (150, 0))],
        )
        _make_glyph(
            font,
            "a_a",
            700,
            contour=False,
            components=[("a", (0, 0)), ("a", (350, 0))],
        )

        font.lib[OPENTYPE_CATEGORIES_KEY] = {"a_a": "base"}
        preliminary = {"a_a": "ligature"}

        glyphSets = [_GlyphSet.from_layer(font)]
        philter = PropagateAnchorsIFilter()
        philter([font], glyphSets, preliminaryOpenTypeCategories=preliminary)

        assert font.lib[OPENTYPE_CATEGORIES_KEY] == {"a_a": "base"}
        anchor_names = {a.name for a in glyphSets[0]["a_a"].anchors}
        assert "top" in anchor_names
        assert "top_1" not in anchor_names
        assert "top_2" not in anchor_names

    def test_depth_sort_uses_component_graphs_from_all_masters(self, FontClass):
        """IFilter ordering must honor deeper component chains in any master."""
        font1 = FontClass()
        _make_glyph(font1, "zz", 350, anchors=[("top", (150, 300))])
        _make_glyph(font1, "mid", 350, anchors=[("top", (150, 300))])
        _make_glyph(
            font1,
            "root",
            350,
            contour=False,
            components=[("mid", (0, 0))],
        )

        font2 = FontClass()
        _make_glyph(font2, "zz", 350, anchors=[("top", (150, 300))])
        _make_glyph(font2, "mid", 350, contour=False, components=[("zz", (0, 0))])
        _make_glyph(
            font2,
            "root",
            350,
            contour=False,
            components=[("mid", (0, 0))],
        )

        glyphSets = [_GlyphSet.from_layer(font1), _GlyphSet.from_layer(font2)]
        philter = PropagateAnchorsIFilter()
        philter([font1, font2], glyphSets)

        assert [(a.name, a.x, a.y) for a in glyphSets[1]["mid"].anchors] == [
            ("top", 150, 300)
        ]
        assert [(a.name, a.x, a.y) for a in glyphSets[1]["root"].anchors] == [
            ("top", 150, 300)
        ]


def test_single_master_preliminary_categories_finalization(FontClass):
    """The single-master filter reads preliminary categories from the private
    lib key (written by the preprocessor) and finalizes after propagation."""
    font = FontClass()
    g = font.newGlyph("a")
    g.width = 350
    pen = g.getPen()
    pen.moveTo((0, 0))
    pen.lineTo((300, 0))
    pen.lineTo((300, 300))
    pen.lineTo((0, 300))
    pen.closePath()
    g.appendAnchor(dict(name="top", x=150, y=300))
    g.appendAnchor(dict(name="bottom", x=150, y=0))

    g = font.newGlyph("acutecomb")
    g.width = 0
    pen = g.getPen()
    pen.moveTo((0, 400))
    pen.lineTo((100, 500))
    pen.lineTo((0, 500))
    pen.closePath()
    g.appendAnchor(dict(name="_top", x=50, y=400))
    g.appendAnchor(dict(name="top", x=50, y=550))

    g = font.newGlyph("a_a")
    g.width = 700
    pen = g.getPen()
    pen.addComponent("a", (1, 0, 0, 1, 0, 0))
    pen.addComponent("a", (1, 0, 0, 1, 350, 0))

    g = font.newGlyph("aacute")
    g.width = 350
    pen = g.getPen()
    pen.addComponent("a", (1, 0, 0, 1, 0, 0))
    pen.addComponent("acutecomb", (1, 0, 0, 1, 0, 0))

    preliminary = {"acutecomb": "mark", "a_a": "ligature"}
    font.lib[_PRELIMINARY_CATEGORIES_KEY] = preliminary

    philter = PropagateAnchorsFilter()
    philter(font)

    assert _PRELIMINARY_CATEGORIES_KEY not in font.lib
    cats = font.lib[OPENTYPE_CATEGORIES_KEY]
    assert cats["acutecomb"] == "mark"
    assert cats["a_a"] == "ligature"
    assert cats["a"] == "base"
    assert cats["aacute"] == "base"

    anchor_names = {a.name for a in font["a_a"].anchors}
    assert "top_1" in anchor_names
    assert "top_2" in anchor_names


def test_single_master_preliminary_ignored_when_source_has_categories(FontClass):
    """Source-level categories take precedence over preliminary in single-master."""
    font = FontClass()
    g = font.newGlyph("a")
    g.width = 350
    pen = g.getPen()
    pen.moveTo((0, 0))
    pen.lineTo((300, 0))
    pen.lineTo((300, 300))
    pen.lineTo((0, 300))
    pen.closePath()
    g.appendAnchor(dict(name="top", x=150, y=300))

    font.lib[OPENTYPE_CATEGORIES_KEY] = {"a": "base"}
    font.lib[_PRELIMINARY_CATEGORIES_KEY] = {"a": "mark"}

    philter = PropagateAnchorsFilter()
    philter(font)

    assert font.lib[OPENTYPE_CATEGORIES_KEY] == {"a": "base"}


class FinalizeCategoriesTest:
    def test_marks_unchanged(self):
        done_anchors = {
            "acutecomb": [AnchorData("_top", 0, 400), AnchorData("top", 0, 550)],
        }
        result = _finalize_categories({"acutecomb": "mark"}, done_anchors)
        assert result["acutecomb"] == "mark"

    def test_ligature_with_anchors_kept(self):
        done_anchors = {
            "f_i": [AnchorData("top_1", 100, 700), AnchorData("top_2", 300, 700)],
        }
        result = _finalize_categories({"f_i": "ligature"}, done_anchors)
        assert result["f_i"] == "ligature"

    def test_ligature_without_anchors_pruned(self):
        done_anchors = {
            "f_i": [],
        }
        result = _finalize_categories({"f_i": "ligature"}, done_anchors)
        assert "f_i" not in result

    def test_ligature_with_only_underscore_anchors_pruned(self):
        done_anchors = {
            "f_i": [AnchorData("_top", 0, 400)],
        }
        result = _finalize_categories({"f_i": "ligature"}, done_anchors)
        assert "f_i" not in result

    def test_base_inferred_from_attaching_anchors(self):
        done_anchors = {
            "a": [AnchorData("top", 150, 300), AnchorData("bottom", 150, 0)],
        }
        result = _finalize_categories({}, done_anchors)
        assert result["a"] == "base"

    def test_no_category_without_attaching_anchors(self):
        done_anchors = {
            "space": [],
        }
        result = _finalize_categories({}, done_anchors)
        assert "space" not in result

    def test_mark_not_reclassified_as_base(self):
        done_anchors = {
            "acutecomb": [AnchorData("_top", 0, 400), AnchorData("top", 0, 550)],
        }
        result = _finalize_categories({"acutecomb": "mark"}, done_anchors)
        assert result["acutecomb"] == "mark"


def test_propagate_anchors_filter_defaults_to_pre():
    """PropagateAnchorsFilter must run before component decomposition."""
    assert PropagateAnchorsFilter().pre is True
    assert PropagateAnchorsIFilter().pre is True


class SparseNestedCompositesTest:
    def test_nested_missing_composites_in_sparse_master(self, FontClass):
        """When a sparse master only has a root composite, and its component
        chain (root -> mid -> base) is entirely missing, anchors should still
        propagate through the interpolated layer.

        Regression test for: the IFilter only resolved one level of missing
        components from the interpolated layer, so nested composites lost
        their anchors.
        """
        # Default master: has all three glyphs
        default_font = FontClass()
        _make_glyph(
            default_font,
            "base",
            600,
            anchors=[("top", (300, 700)), ("bottom", (300, 0))],
        )
        _make_glyph(
            default_font,
            "mid",
            600,
            components=[("base", (0, 0))],
        )
        _make_glyph(
            default_font,
            "root",
            600,
            components=[("mid", (0, 0))],
        )

        # Sparse master: only has root
        sparse_font = FontClass()
        _make_glyph(
            sparse_font,
            "root",
            600,
            components=[("mid", (0, 0))],
        )

        ds = DesignSpaceDocument()
        a = AxisDescriptor()
        a.tag = "wght"
        a.name = "Weight"
        a.minimum = a.default = 400
        a.maximum = 700
        ds.addAxis(a)

        s1 = SourceDescriptor()
        s1.font = default_font
        s1.location = {"Weight": 400}
        ds.addSource(s1)

        s2 = SourceDescriptor()
        s2.font = sparse_font
        s2.location = {"Weight": 700}
        ds.addSource(s2)

        glyphSets = [
            _GlyphSet.from_layer(default_font),
            _GlyphSet.from_layer(sparse_font),
        ]

        instantiator = Instantiator.from_designspace(
            ds, do_kerning=False, do_info=False
        )
        philter = PropagateAnchorsIFilter()
        philter([default_font, sparse_font], glyphSets, instantiator)

        # Default master: straightforward propagation
        assert {a.name for a in glyphSets[0]["root"].anchors} == {"top", "bottom"}

        # Sparse master: root should still get anchors via interpolated mid -> base
        assert {a.name for a in glyphSets[1]["root"].anchors} == {"top", "bottom"}



"""
InfoCompiler is used to apply fontinfo overrides to an already compiled font.
This is used to apply fontinfo from a DesignSpace variable-font after merging
font sources into final variable font.

It builds a temporary font with the only the tables that can be modified with
fontinfo, then merge relevant table attributes it into the original font.
"""

import copy

from ufo2ft.outlineCompiler import BaseOutlineCompiler


class InfoCompiler(BaseOutlineCompiler):
    info_tables = frozenset(
        [
            "head",
            "hhea",
            "name",
            "OS/2",
            "post",
            "vhea",
            "gasp",
        ]
    )

    def __init__(self, otf, ufo, info):
        self.orig_otf = otf
        tables = self.info_tables & set(otf.keys())

        # Create a temporary UFO and sets its fontinfo to the union of the main
        # UFO's fontinfo and the DesignSpace variable-font’s info.
        temp_ufo = type(ufo)()
        if hasattr(ufo.info, "getDataForSerialization"):
            # defcon
            data = ufo.info.getDataForSerialization()
            data.update(info)
            temp_ufo.info.setDataFromSerialization(data)
        else:
            # ufoLib2
            temp_ufo.info = copy.copy(ufo.info)
            for k, v in info.items():
                setattr(temp_ufo.info, k, v)

        super().__init__(temp_ufo, tables=tables, glyphSet={}, glyphOrder=[])

    def compile(self):
        super().compile()
        if "gasp" in self.tables:
            self.setupTable_gasp()
        return self.orig_otf

    @staticmethod
    def makeMissingRequiredGlyphs(*args, **kwargs):
        return

    def makeFontBoundingBox(self):
        from ufo2ft.outlineCompiler import EMPTY_BOUNDING_BOX

        return EMPTY_BOUNDING_BOX

    def _set_attrs(self, tag, attrs):
        temp = self.otf[tag]
        orig = self.orig_otf[tag]
        for attr in attrs:
            if (value := getattr(temp, attr, None)) is not None:
                setattr(orig, attr, value)

    def setupTable_head(self):
        super().setupTable_head()
        self._set_attrs(
            "head",
            {
                "fontRevision",
                "unitsPerEm",
                "created",
                "macStyle",
                "flags",
                "lowestRecPPEM",
            },
        )

    def setupTable_hhea(self):
        super().setupTable_hhea()
        self._set_attrs(
            "hhea",
            {
                "ascent",
                "descent",
                "lineGap",
                "caretSlopeRise",
                "caretSlopeRun",
                "caretOffset",
            },
        )

    def setupTable_vhea(self):
        super().setupTable_vhea()
        self._set_attrs(
            "vhea",
            {
                "ascent",
                "descent",
                "lineGap",
                "caretSlopeRise",
                "caretSlopeRun",
                "caretOffset",
            },
        )

    def setupTable_OS2(self):
        super().setupTable_OS2()
        self._set_attrs(
            "OS/2",
            {
                "usWeightClass",
                "usWidthClass",
                "fsType",
                "ySubscriptXSize",
                "ySubscriptYSize",
                "ySubscriptYOffset",
                "ySubscriptXOffset",
                "ySuperscriptXSize",
                "ySuperscriptYSize",
                "ySuperscriptYOffset",
                "ySuperscriptXOffset",
                "yStrikeoutSize",
                "yStrikeoutPosition",
                "sFamilyClass",
                "panose",
                "ulUnicodeRange1",
                "ulUnicodeRange2",
                "ulUnicodeRange3",
                "ulUnicodeRange4",
                "achVendID",
                "fsSelection",
                "sTypoAscender",
                "sTypoDescender",
                "sTypoLineGap",
                "usWinAscent",
                "usWinDescent",
                "ulCodePageRange1",
                "ulCodePageRange2",
                "sxHeight",
                "sCapHeight",
            },
        )

    def setupTable_post(self):
        super().setupTable_post()
        self._set_attrs(
            "post",
            {
                "italicAngle",
                "underlinePosition",
                "underlineThickness",
                "isFixedPitch",
            },
        )

    def setupTable_name(self):
        super().setupTable_name()
        temp = self.otf["name"]
        orig = self.orig_otf["name"]
        temp_names = {
            (n.nameID, n.platformID, n.platEncID, n.langID): n for n in temp.names
        }
        orig_names = {
            (n.nameID, n.platformID, n.platEncID, n.langID): n for n in orig.names
        }
        orig_names.update(temp_names)
        orig.names = list(orig_names.values())
        # If the 'typographic' family/subfamily names were present in the original
        # font's name table (as built from the base master UFO's fontinfo.plist) but
        # are now omitted from the 'temp' font's (built from the VF's public.fontInfo)
        # it means they are no longer needed after merging the latter into the former
        # and should be removed.
        typo_name_ids = {16, 17}
        if typo_name_ids.issubset(
            {n.nameID for n in orig.names}
        ) and not typo_name_ids.issubset({n.nameID for n in temp.names}):
            orig.removeNames(nameID=16)
            orig.removeNames(nameID=17)
            # Any references to the removed typographic style name (nameID 17) in either
            # fvar instances or STAT axis values must be replaced with the legacy style
            # name (nameID 2).
            if "fvar" in self.orig_otf:
                for inst in self.orig_otf["fvar"].instances:
                    if inst.subfamilyNameID == 17:
                        inst.subfamilyNameID = 2
            if "STAT" in self.orig_otf:
                stat = self.orig_otf["STAT"].table
                if stat.ElidedFallbackNameID == 17:
                    stat.ElidedFallbackNameID = 2
                if stat.AxisValueArray:
                    for rec in stat.AxisValueArray.AxisValue:
                        if rec.ValueNameID == 17:
                            rec.ValueNameID = 2

    def setupTable_gasp(self):
        from ufo2ft.instructionCompiler import InstructionCompiler

        instructionCompiler = InstructionCompiler(self.ufo, self.otf)
        instructionCompiler.setupTable_gasp()
        self._set_attrs("gasp", {"gaspRange"})

    def setupTable_maxp(self):
        return



from enum import IntEnum
from types import MappingProxyType


class CFFOptimization(IntEnum):
    NONE = 0
    SPECIALIZE = 1
    SUBROUTINIZE = 2


SPARSE_TTF_MASTER_TABLES = frozenset(
    ["glyf", "head", "hmtx", "loca", "maxp", "post", "vmtx", "cvt ", "fpgm", "prep"]
)
SPARSE_OTF_MASTER_TABLES = frozenset(["CFF ", "VORG", "head", "hmtx", "maxp", "vmtx"])

UFO2FT_PREFIX = "com.github.googlei18n.ufo2ft."
GLYPHS_PREFIX = "com.schriftgestaltung."

FILTERS_KEY = UFO2FT_PREFIX + "filters"

EXPLICIT_CLOSING_LINE_KEY = UFO2FT_PREFIX + "explicitClosingLine"

MTI_FEATURES_PREFIX = UFO2FT_PREFIX + "mtiFeatures"

FEATURE_WRITERS_KEY = UFO2FT_PREFIX + "featureWriters"

USE_PRODUCTION_NAMES = UFO2FT_PREFIX + "useProductionNames"
GLYPHS_DONT_USE_PRODUCTION_NAMES = GLYPHS_PREFIX + "Don't use Production Names"
GLYPHS_COMPONENT_INFO_KEY = GLYPHS_PREFIX + "Glyphs.ComponentInfo"
KEEP_GLYPH_NAMES = UFO2FT_PREFIX + "keepGlyphNames"

COLOR_LAYERS_KEY = UFO2FT_PREFIX + "colorLayers"
COLOR_PALETTES_KEY = UFO2FT_PREFIX + "colorPalettes"
COLOR_LAYER_MAPPING_KEY = UFO2FT_PREFIX + "colorLayerMapping"
# sequence of [glyphs, clipBox], where 'glyphs' is in turn a sequence of
# glyph names, and 'clipBox' a 5- or 4-item sequence of numbers:
# Sequence[
#   Sequence[
#     Sequence[str, ...],  # glyph names
#     Union[
#       Sequence[float, float, float, float, float],  # variable box
#       Sequence[float, float, float, float],  # non-variable box
#     ]
#   ],
#   ...
# ]
COLR_CLIP_BOXES_KEY = UFO2FT_PREFIX + "colrClipBoxes"

GLYPHS_MATH_PREFIX = "com.nagwa.MATHPlugin."
GLYPHS_MATH_CONSTANTS_KEY = GLYPHS_MATH_PREFIX + "constants"
GLYPHS_MATH_VARIANTS_KEY = GLYPHS_MATH_PREFIX + "variants"
GLYPHS_MATH_EXTENDED_SHAPE_KEY = GLYPHS_MATH_PREFIX + "extendedShape"

OBJECT_LIBS_KEY = "public.objectLibs"
OPENTYPE_CATEGORIES_KEY = "public.openTypeCategories"
_PRELIMINARY_CATEGORIES_KEY = UFO2FT_PREFIX + "__preliminaryOpenTypeCategories"
OPENTYPE_META_KEY = "public.openTypeMeta"
OPENTYPE_POST_UNDERLINE_POSITION_KEY = "public.openTypePostUnderlinePosition"
TRUETYPE_INSTRUCTIONS_KEY = "public.truetype.instructions"
TRUETYPE_METRICS_KEY = "public.truetype.useMyMetrics"
TRUETYPE_OVERLAP_KEY = "public.truetype.overlap"
TRUETYPE_ROUND_KEY = "public.truetype.roundOffsetToGrid"
UNICODE_VARIATION_SEQUENCES_KEY = "public.unicodeVariationSequences"

COMMON_SCRIPT = "Zyyy"

UNICODE_SCRIPT_ALIASES = MappingProxyType({"Hira": "Hrkt", "Kana": "Hrkt"})


# HarfBuzz passes Sinhala to the Indic shaper, while OpenType moved it to the USE shaper.

INDIC_SCRIPTS = [
    "Beng",  # Bengali
    "Deva",  # Devanagari
    "Gujr",  # Gujarati
    "Guru",  # Gurmukhi
    "Knda",  # Kannada
    "Mlym",  # Malayalam
    "Orya",  # Oriya
    "Sinh",  # Sinhala
    "Taml",  # Tamil
    "Telu",  # Telugu
]

USE_SCRIPTS = [
    # Correct as at Unicode 15.0
    "Adlm",  # Adlam
    "Ahom",  # Ahom
    "Bali",  # Balinese
    "Batk",  # Batak
    "Brah",  # Brahmi
    "Bugi",  # Buginese
    "Buhd",  # Buhid
    "Cakm",  # Chakma
    "Cham",  # Cham
    "Chrs",  # Chorasmian
    "Cpmn",  # Cypro Minoan
    "Diak",  # Dives Akuru
    "Dogr",  # Dogra
    "Dupl",  # Duployan
    "Egyp",  # Egyptian Hieroglyphs
    "Elym",  # Elymaic
    "Gong",  # Gunjala Gondi
    "Gonm",  # Masaram Gondi
    "Gran",  # Grantha
    "Hano",  # Hanunoo
    "Hmng",  # Pahawh Hmong
    "Hmnp",  # Nyiakeng Puachue Hmong
    "Java",  # Javanese
    "Kali",  # Kayah Li
    "Kawi",  # Kawi
    "Khar",  # Kharosthi
    "Khoj",  # Khojki
    "Kits",  # Khitan Small Script
    "Kthi",  # Kaithi
    "Lana",  # Tai Tham
    "Lepc",  # Lepcha
    "Limb",  # Limbu
    "Mahj",  # Mahajani
    "Maka",  # Makasar
    "Mand",  # Mandaic
    "Mani",  # Manichaean
    "Marc",  # Marchen
    "Medf",  # Medefaidrin
    "Modi",  # Modi
    "Mong",  # Mongolian
    "Mtei",  # Meetei Mayek
    "Mult",  # Multani
    "Nagm",  # Nag Mundari
    "Nand",  # Nandinagari
    "Newa",  # Newa
    "Bhks",  # Bhaiksuki
    "Nko ",  # Nko  (this one is incorrect but probably won't hurt to keep?)
    "Nkoo",  # Nko
    "Ougr",  # Old Uyghur
    "Phag",  # Phags Pa
    "Phlp",  # Psalter Pahlavi
    "Plrd",  # Miao
    "Rjng",  # Rejang
    "Rohg",  # Hanifi Rohingya
    "Saur",  # Saurashtra
    "Shrd",  # Sharada
    "Sidd",  # Siddham
    "Sind",  # Khudawadi
    "Sogd",  # Sogdian
    "Sogo",  # Old Sogdian
    "Soyo",  # Soyombo
    "Sund",  # Sundanese
    "Sylo",  # Syloti Nagri
    "Tagb",  # Tagbanwa
    "Takr",  # Takri
    "Tale",  # Tai Le
    "Tavt",  # Tai Viet
    "Tfng",  # Tifinagh
    "Tglg",  # Tagalog
    "Tibt",  # Tibetan
    "Tirh",  # Tirhuta
    "Tnsa",  # Tangsa
    "Toto",  # Toto
    "Vith",  # Vithkuqi
    "Wcho",  # Wancho
    "Yezi",  # Yezidi
    "Zanb",  # Zanabazar Square
]

ANCHOR_LIB_GPOS_CONTEXT_KEY = "GPOS_Context"



import logging
from enum import Enum

from ufo2ft.filters import BaseFilter

logger = logging.getLogger(__name__)


class RemoveOverlapsFilter(BaseFilter):
    class Backend(Enum):
        BOOLEAN_OPERATIONS = "booleanOperations"
        SKIA_PATHOPS = "pathops"

    # use booleanOperations by default, unless pathops specified as backend
    _kwargs = {"backend": Backend.BOOLEAN_OPERATIONS}

    def start(self):
        self.options.backend = self.Backend(self.options.backend)

        if self.options.backend is self.Backend.BOOLEAN_OPERATIONS:
            from booleanOperations import BooleanOperationsError, union

            self.union = union
            self.Error = BooleanOperationsError
            self.penGetter = "getPointPen"

            logger.debug("using booleanOperations as RemoveOverlapsFilter backend")
        elif self.options.backend is self.Backend.SKIA_PATHOPS:
            from pathops import PathOpsError, union

            self.union = union
            self.Error = PathOpsError
            self.penGetter = "getPen"

            logger.debug("using skia-pathops as RemoveOverlapsFilter backend")
        else:
            raise AssertionError(self.options.backend)

    def filter(self, glyph):
        if not len(glyph):
            return False

        contours = list(glyph)
        glyph.clearContours()
        pen = getattr(glyph, self.penGetter)()
        try:
            self.union(contours, pen)
        except (self.Error, ValueError, TypeError):
            logger.error("Failed to remove overlaps for %s", glyph.name)
            raise
        return True



from __future__ import annotations

from typing import TYPE_CHECKING

from fontTools.pens.recordingPen import RecordingPointPen

from ufo2ft.constants import EXPLICIT_CLOSING_LINE_KEY
from ufo2ft.filters import BaseIFilter

if TYPE_CHECKING:
    from ufoLib2.objects import Glyph


class ExplicitClosingLineIFilter(BaseIFilter):
    """Mark CFF glyphs that need compatible explicit closing lines."""

    _pre = True

    def filter(self, glyphName: str, glyphs: list[Glyph]) -> bool:
        if len(glyphs) < 2:
            return False

        needs_explicit_closing = _needs_explicit_closing_lines(glyphs)
        modified = False
        for glyph in glyphs:
            if needs_explicit_closing:
                if glyph.lib.get(EXPLICIT_CLOSING_LINE_KEY) is not True:
                    glyph.lib[EXPLICIT_CLOSING_LINE_KEY] = True
                    modified = True
            elif glyph.lib.pop(EXPLICIT_CLOSING_LINE_KEY, None) is not None:
                modified = True
        return modified


def _needs_explicit_closing_lines(glyphs) -> bool:
    contours = [_get_contours(glyph) for glyph in glyphs]
    if not contours:
        return False

    num_contours = len(contours[0])
    if not num_contours or any(
        len(glyph_contours) != num_contours for glyph_contours in contours
    ):
        return False

    for contour_index in range(num_contours):
        contour_group = [glyph_contours[contour_index] for glyph_contours in contours]
        point_types = [
            tuple(point_type for _, point_type in contour) for contour in contour_group
        ]
        if any(types != point_types[0] for types in point_types):
            continue

        explicit_closing = [
            _has_explicit_closing_line(contour) for contour in contour_group
        ]
        if any(explicit_closing) and not all(explicit_closing):
            return True

    return False


def _get_contours(glyph):
    pen = RecordingPointPen()
    glyph.drawPoints(pen)

    contours = []
    current_contour = None
    for operator, args, _kwargs in pen.value:
        if operator == "beginPath":
            current_contour = []
        elif operator == "addPoint":
            if current_contour is None:
                continue
            pt, segment_type = args[:2]
            current_contour.append((pt, segment_type))
        elif operator == "endPath":
            if current_contour is not None:
                contours.append(current_contour)
                current_contour = None

    return contours


def _first_oncurve_index(contour):
    for i, (_, segment_type) in enumerate(contour):
        if segment_type is not None:
            return i
    return None


def _has_explicit_closing_line(contour) -> bool:
    if not contour or contour[0][1] == "move":
        return False

    first_oncurve = _first_oncurve_index(contour)
    if first_oncurve is None:
        return False

    first_point, first_type = contour[first_oncurve]
    if first_type != "line":
        return False

    previous_point, _ = contour[first_oncurve - 1]
    return first_point == previous_point
